import binascii
import struct
from .util import ofpp, parseInt, get_token, parse_func
from .oxm import str2oxm, oxm2str
from .nx import nxast, register_nxast, NX_VENDOR_ID

align8 = lambda x:(x+7)//8*8

OFPT_FLOW_MOD = 14
OFPT_GROUP_MOD = 15
OFPT_MULTIPART_REQUEST = 18
OFPT_MULTIPART_REPLY = 19

OFPMP_FLOW = 1
OFPMP_GROUP_DESC = 7

OFPTT_MAX = 0xfe
OFPTT_ALL = 0xff

OFPCML_MAX = 0xffe5
OFPCML_NO_BUFFER = 0xffff

OFP_NO_BUFFER = 0xffffffff

for num,name in ofpp.items():
	globals()["OFPP_{:s}".format(name.upper())] = num

ofpg = {
	0xffffff00: "max",
	0xfffffffc: "all",
	0xffffffff: "any",
}
for num,name in ofpg.items():
	globals()["OFPG_{:s}".format(name.upper())] = num

ofpff = {
	"send_flow_rem": 1<<0,
	"check_overlap": 1<<1,
	"reset_counts": 1<<2,
	"no_pkt_counts": 1<<3,
	"no_byt_counts": 1<<4 }

for name,num in ofpff.items():
	globals()["OFPFF_{:s}".format(name.upper())] = num

OFPFC_ADD = 0
OFPFC_MODIFY = 1
OFPFC_MODIFY_STRICT = 2
OFPFC_DELETE = 3
OFPFC_DELETE_STRICT = 4

OFPMT_OXM = 1

action_names = {
	0: "output",
	11: "copy_ttl_out",
	12: "copy_ttl_in",
	15: "set_mpls_ttl",
	16: "dec_mpls_ttl",
	17: "push_vlan",
	18: "pop_vlan",
	19: "push_mpls",
	20: "pop_mpls",
	21: "set_queue",
	22: "group",
	23: "set_nw_ttl",
	24: "dec_nw_ttl",
	25: "set_field",
	26: "push_pbb",
	27: "pop_pbb",
	0xffff: "experimenter",
}
for n,name in action_names.items():
	globals()["OFPAT_{:s}".format(name.upper())] = n

def action_generic_str2act(ofpat):
	def str2act(payload, readarg):
		return struct.pack("!HH4x", ofpat, 8), 0
	return str2act

def action_generic_act2str(name):
	def act2str(payload):
		assert payload == bytearray(4)
		return name
	return act2str

def action_uint_str2act(ofpat, pack):
	def str2act(unparsed, readarg):
		h,b,unparsed = get_token(unparsed)
		num,l = parseInt(b)
		assert l==len(b)
		value = struct.pack(pack, num)
		return struct.pack("!HH", ofpat, 4+len(value))+value, len(h)+len(b)

	return str2act

def action_uint_act2str(fmt, pack):
	def act2str(payload):
		return fmt.format(*struct.unpack_from(pack, payload))
	return act2str

def action_output_str2act(unparsed, readarg):
	h,b,unparsed = get_token(unparsed)
	assert b.find("/")<0
	ps = b.split(":", 2)
	port = None
	for num,name in ofpp.items():
		if name == ps[0]:
			port = num
	if port is None:
		port,l = parseInt(ps[0])
		assert len(ps[0]) == l

	maxLen = OFPCML_NO_BUFFER
	if len(ps) > 1:
		maxLen,l = parseInt(ps[1])
		assert len(ps[1]) == l

	return struct.pack("!HHIH6x", OFPAT_OUTPUT, 16, port, maxLen), len(h) + len(b)

def action_output_act2str(payload):
	(port,maxLen) = struct.unpack("!IH6x", payload)
	name = ofpp.get(port, "{:d}".format(port))
	if port == OFPP_CONTROLLER and maxLen != OFPCML_NO_BUFFER:
		return "output={:s}:{:#x}".format(name, maxLen)
	else:
		return "output={:s}".format(name)

def action_push_str2act(ofpat):
	def str2act(unparsed, readarg):
		h,b,unparsed = get_token(unparsed)
		num, l = parseInt(b)
		return struct.pack("!HHH2x", ofpat, 8, num), len(h)+len(b)
	return str2act

def action_push_act2str(name):
	def act2str(payload):
		num = struct.unpack_from("!H", payload)[0]
		return "{:s}={:#06x}".format(name, num)
	return act2str

_str2act = {}
_act2str = {}

_str2act["output"] = action_output_str2act
_act2str[OFPAT_OUTPUT] = action_output_act2str

_str2act["copy_ttl_out"] = action_generic_str2act(OFPAT_COPY_TTL_OUT)
_act2str[OFPAT_COPY_TTL_OUT] = action_generic_act2str("copy_ttl_out")

_str2act["copy_ttl_in"] = action_generic_str2act(OFPAT_COPY_TTL_IN)
_act2str[OFPAT_COPY_TTL_IN] = action_generic_act2str("copy_ttl_in")

_str2act["set_mpls_ttl"] = action_uint_str2act(OFPAT_SET_MPLS_TTL, "!B3x")
_act2str[OFPAT_SET_MPLS_TTL] = action_uint_act2str("set_mpls_ttl={:d}", "!B3x")

_str2act["dec_mpls_ttl"] = action_generic_str2act(OFPAT_DEC_MPLS_TTL)
_act2str[OFPAT_DEC_MPLS_TTL] = action_generic_act2str("dec_mpls_ttl")

_str2act["push_vlan"] = action_push_str2act(OFPAT_PUSH_VLAN)
_act2str[OFPAT_PUSH_VLAN] = action_push_act2str("push_vlan")

_str2act["pop_vlan"] = action_generic_str2act(OFPAT_POP_VLAN)
_act2str[OFPAT_POP_VLAN] = action_generic_act2str("pop_vlan")

_str2act["push_mpls"] = action_push_str2act(OFPAT_PUSH_MPLS)
_act2str[OFPAT_PUSH_MPLS] = action_push_act2str("push_mpls")

_str2act["pop_mpls"] = action_push_str2act(OFPAT_POP_MPLS)
_act2str[OFPAT_POP_MPLS] = action_push_act2str("pop_mpls")

_str2act["push_mpls"] = action_push_str2act(OFPAT_PUSH_MPLS)
_act2str[OFPAT_PUSH_MPLS] = action_push_act2str("push_mpls")

_str2act["set_queue"] = action_uint_str2act(OFPAT_SET_QUEUE, "!I")
_act2str[OFPAT_SET_QUEUE] = action_uint_act2str("set_queue={:d}", "!I")

_str2act["group"] = action_uint_str2act(OFPAT_GROUP, "!I")
_act2str[OFPAT_GROUP] = action_uint_act2str("group={:d}", "!I")

_str2act["set_nw_ttl"] = action_uint_str2act(OFPAT_SET_NW_TTL, "!B")
_act2str[OFPAT_SET_NW_TTL] = action_uint_act2str("set_nw_ttl={:d}", "!B")

_str2act["dec_nw_ttl"] = action_generic_str2act(OFPAT_DEC_NW_TTL)
_act2str[OFPAT_DEC_NW_TTL] = action_generic_act2str("dec_nw_ttl")

_str2act["push_pbb"] = action_push_str2act(OFPAT_PUSH_PBB)
_act2str[OFPAT_PUSH_PBB] = action_push_act2str("push_pbb")

_str2act["pop_pbb"] = action_generic_str2act(OFPAT_POP_PBB)
_act2str[OFPAT_POP_PBB] = action_generic_act2str("pop_pbb")

register_nxast(_str2act, _act2str)

def str2act(s):
	h,name,arg = get_token(s)
	fname,farg = parse_func(name)

	if farg and fname in _str2act:
		b,p = _str2act[fname](farg, False)
		return bytes(b), len(h)+len(name)
	elif name in _str2act:
		b,p = _str2act[name](arg, True)
		return bytes(b), len(h)+len(name)+p
	elif name.startswith("set_"):
		if farg:
			op,payload,s = get_token(farg)
			oxm,p = str2oxm(fname[4:]+"="+payload, loop=False)
			consumed_length = len(h)+len(name)
		else:
			op,payload,s = get_token(arg)
			oxm,p = str2oxm(name[4:]+op+payload, loop=False)
			consumed_length = len(h)+4+p

		l = align8(len(oxm)+4)
		ret = bytearray(l)
		ret[:4] = struct.pack("!HH", OFPAT_SET_FIELD, l)
		ret[4:4+len(oxm)] = oxm
		return bytes(ret), consumed_length
	else:
		return b"", len(h)

def act2str(msg, loop=True):
	tokens = []
	while len(msg) > 4:
		(atype,l) = struct.unpack_from("!HH", msg)
		offset = 4
		if atype == OFPAT_EXPERIMENTER:
			vendor = struct.unpack_from("!I", msg, 4)[0]
			if vendor == NX_VENDOR_ID:
				atype = nxast(struct.unpack_from("!H", msg, 8)[0])
				offset = 10

		act = _act2str.get(atype)
		if atype == OFPAT_SET_FIELD:
			tokens.append("set_"+oxm2str(msg[4:], loop=False))
		elif act:
			tokens.append(act(msg[offset:l]))
		else:
			tokens.append("?")

		if loop:
			msg = msg[l:]
		else:
			break

	return ",".join(tokens)

instruction_names = {
	1: "goto_table",
	2: "write_metadata",
	3: "write_actions",
	4: "apply_actions",
	5: "clear_actions",
	6: "meter"
}
for n,name in instruction_names.items():
	globals()["OFPIT_{:s}".format(name.upper())] = n


def inst2str(msg, loop=True):
	tokens = []
	while len(msg) > 4:
		(itype,l) = struct.unpack_from("!HH", msg)
		if itype == OFPIT_GOTO_TABLE:
			assert l==8
			tokens.append("@goto={:d}".format(*struct.unpack_from("!B", msg, 4)))
		elif itype == OFPIT_WRITE_METADATA:
			assert l==24
			(v,m) = struct.unpack_from("!QQ", msg, 8)
			if m == 0:
				tokens.append("@metadata={:#x}".format(v))
			else:
				tokens.append("@metadata={:#x}/{:#x}".format(v,m))
		elif itype == OFPIT_WRITE_ACTIONS:
			assert l%8==0
			tokens.append("@write")
			arg = act2str(msg[8:l])
			if len(arg):
				tokens.append(arg)
		elif itype == OFPIT_APPLY_ACTIONS:
			assert l%8==0
			tokens.append("@apply")
			arg = act2str(msg[8:l])
			if len(arg):
				tokens.append(arg)
		elif itype == OFPIT_CLEAR_ACTIONS:
			assert l == 8
			tokens.append("@clear")
		elif itype == OFPIT_METER:
			assert l == 8, repr(msg)
			tokens.append("@meter={:d}".format(*struct.unpack_from("!I", msg, 4)))
		else:
			tokens.append("?")

		if loop:
			msg = msg[l:]
		else:
			break

	return ",".join(tokens)

PHASE_MATCH = 0
PHASE_ACTION = 1
PHASE_NOARG = 2

def str2dict(s):
	ret = dict(
		match= b"",
		inst= b"",
	)

	actions = b""
	def inst_action(atype):
		def func():
			ret["inst"] += struct.pack("!HH4x", atype, 8+len(actions))+actions
		return func

	func = None
	phase = PHASE_MATCH
	while len(s) > 0:
		h,name,s = get_token(s)
		assert h.find("=")<0
		if name.startswith("@"):
			if func is not None:
				func()

			func = None
			phase = PHASE_NOARG
			if name in ("@goto", "@goto_table"):
				op,payload,s = get_token(s)
				assert op.find("=")>=0, "goto requires arg"
				num,l = parseInt(payload)
				assert l == len(payload)
				ret["inst"] += struct.pack("!HHB3x", OFPIT_GOTO_TABLE, 8, num)
			elif name in ("@metadata", "@write_metadata"):
				op,payload,s = get_token(s)
				assert op.find("=")>=0, "metadata requires arg"
				vm = payload.split("/", 1)
				num,l = parseInt(vm[0])
				assert l == len(vm[0])
				if len(vm) > 1:
					mask,l = parseInt(vm[1])
					assert l == len(vm[1])
					ret["inst"] += struct.pack("!HH4xQQ", OFPIT_WRITE_METADATA, 24, num, mask)
				else:
					ret["inst"] += struct.pack("!HH4xQQ", OFPIT_WRITE_METADATA, 24, num, 0)
			elif name in ("@apply", "@apply_actions"):
				func = inst_action(OFPIT_APPLY_ACTIONS)
				actions = b""
				phase = PHASE_ACTION
			elif name in ("@write", "@write_actions"):
				func = inst_action(OFPIT_WRITE_ACTIONS)
				actions = b""
				phase = PHASE_ACTION
			elif name in ("@clear", "@clear_actions"):
				ret["inst"] += struct.pack("!HH4x", OFPIT_CLEAR_ACTIONS, 8)
			elif name == "@meter":
				op,payload,s = get_token(s)
				assert op.find("=")>=0, "meter requires arg"
				assert payload.find("/") < 0, "meter does not take mask"
				num,l = parseInt(payload)
				assert l == len(payload)
				ret["inst"] += struct.pack("!HHI", OFPIT_METER, 8, num)
			else:
				raise ValueError("unknown {:s}".format(name))
		elif phase == PHASE_MATCH:
			def proc(field):
				op,payload,unparsed = get_token(s)
				assert op.find("=")>=0 and payload.find("/")<0
				num,l = parseInt(payload)
				assert len(payload) == l
				ret[field] = num
				return unparsed
			if name in ("table", "priority", "idle_timeout", "hard_timeout", "buffer"):
				s = proc(name)
			elif name in ofpff:
				ret["flags"] = ret.get("flags", 0) | ofpff[name]
			elif name == "cookie":
				op,payload,s = get_token(s)
				assert op.find("=")>=0, "cookie take value"
				vm = payload.split("/", 1)
				num,l = parseInt(vm[0])
				assert len(vm[0]) == l
				ret[name] = num
				if len(vm) > 1:
					num,l = parseInt(vm[1])
					ret["cookie_mask"] = num
			elif name == "out_port":
				op,payload,s = get_token(s)
				assert op.find("=") >= 0 and payload.find("/") < 0
				port = None
				for num,pname in ofpp.items():
					if pname == payload:
						port = num
				if port is None:
					port,l = parseInt(payload)
					assert l == len(payload)
				ret[name] = port
			elif name == "out_group":
				op,payload,s = get_token(s)
				assert op.find("=")>=0 and payload.find("/")<0
				port = None
				for num,gname in ofpg.items():
					if gname == payload:
						port = num
				if port is None:
					port,l = parseInt(payload)
					assert l == len(payload)
				ret[name] = port
			else:
				oxm, l = str2oxm(name+s, loop=False)
				if l == 0:
					raise ValueError("unknown match {:s}".format(s))
				ret["match"] += oxm
				s = (name+s)[l:]
		elif phase == PHASE_ACTION:
			act, l = str2act(name+s)
			if l == 0:
				raise ValueError("unknown action {:s}".format(s))
			actions += act
			s = (name+s)[l:]
		else:
			raise ValueError("invalid syntax")
	if func:
		func()

	return ret

def _fixed(default, **kwargs):
	'''
	@param default parameters the is equal in this will be suppressed
	'''
	def emit(name):
		return name in kwargs and kwargs[name] != default.get(name)
	
	ret = []
	if emit("cookie_mask"):
		ret.append("cookie={:#x}/{:#x}".format(kwargs["cookie"], kwargs["cookie_mask"]))
	elif emit("cookie"):
		ret.append("cookie={:#x}".format(kwargs["cookie"]))
	
	if emit("table"):
		ret.append("table={:d}".format(kwargs["table"]))

	if emit("priority"):
		ret.append("priority={:d}".format(kwargs["priority"]))

	if emit("buffer"):
		ret.append("buffer={:#x}".format(kwargs["buffer"]))

	if emit("out_port"):
		out_port = kwargs["out_port"]
		if out_port in ofpp:
			ret.append("out_port={:s}".format(ofpp[out_port]))
		else:
			ret.append("out_port={:d}".format(out_port))

	if emit("out_group"):
		out_group = kwargs["out_group"]
		if out_group in ofpg:
			ret.append("out_group={:s}".format(ofpg[out_group]))
		else:
			ret.append("out_group={:d}".format(out_group))

	if emit("idle_timeout"):
		ret.append("idle_timeout={:d}".format(kwargs["idle_timeout"]))

	if emit("hard_timeout"):
		ret.append("hard_timeout={:d}".format(kwargs["hard_timeout"]))
	
	if emit("flags"):
		for name,num in ofpff.items():
			if kwargs["flags"] & num:
				ret.append(name)
	
	return ret

ofpgc = ("add", "modify", "delete")
for num, name in enumerate(ofpgc):
	globals()["OFPGC_{:s}".format(name.upper())] = num

ofpgt = ("all", "select", "indirect", "ff")
for num, name in enumerate(ofpgt):
	globals()["OFPGT_{:s}".format(name.upper())] = num

def buckets2str(msg, group_type=OFPGT_ALL):
	buckets = []
	while len(msg) >= 16:
		(length,
		weight,
		watch_port,
		watch_group) = struct.unpack_from("!HHII4x", msg)
		ret = dict(actions=act2str(msg[16:length]))
		if group_type == OFPGT_SELECT:
			ret["weight"] = weight
		if group_type == OFPGT_FF:
			if watch_port != OFPP_ANY:
				ret["watch_port"] = watch_port
			if watch_group != OFPG_ANY:
				ret["watch_group"] = watch_group
		buckets.append(ret)
		msg = msg[length:]
	return buckets

def str2buckets(buckets, group_type=OFPGT_ALL):
	ret = b""
	for b in buckets:
		acts,_ = str2act(b.get("actions", ""))
		ret += struct.pack("!HHII4x",
			align8(16+len(acts)),
			b.get("weight", 0),
			b.get("watch_port", OFPP_ANY),
			b.get("watch_group", OFPG_ANY))
		ret += acts
		ret += b"\0"*(align8(len(acts))-len(acts))
	return ret

ofpfc_del_default = dict(
	cookie = 0,
	cookie_mask = 0,
	table = OFPTT_ALL,
	idle_timeout = 0,
	hard_timeout = 0,
	priority = 0x8000,
	buffer = 0,
	out_port = OFPP_ANY,
	out_group = OFPG_ANY,
	flags = 0,
)

ofpfc_default = dict(
	cookie = 0,
	cookie_mask = 0,
	table = 0,
	idle_timeout = 0,
	hard_timeout = 0,
	priority = 0x8000,
	buffer = OFP_NO_BUFFER,
	out_port = 0,
	out_group = 0,
	flags = 0,
)

def str2mod(s, command=OFPFC_ADD, xid=0):
	default = ofpfc_default
	if command in (OFPFC_DELETE, OFPFC_DELETE_STRICT):
		default = ofpfc_del_default

	info = dict(default)
	info.update(str2dict(s))

	OFPMT_OXM = 1
	oxm = info.get("match", b"")
	length = 4 + len(oxm)
	match = struct.pack("!HH", OFPMT_OXM, length) + oxm
	match += b"\0" * (align8(length)-length)

	inst = info.get("inst", b"")

	return struct.pack("!BBHIQQBBHHHIIIH2x", 4, OFPT_FLOW_MOD, 48+align8(length)+len(inst), xid,
		info["cookie"],
		info["cookie_mask"],
		info["table"],
		command,
		info["idle_timeout"],
		info["hard_timeout"],
		info["priority"],
		info["buffer"],
		info["out_port"],
		info["out_group"],
		info["flags"])+match+inst

def mod2str(msg):
	(hdr_version, hdr_type, hdr_length, hdr_xid,
	cookie,
	cookie_mask,
	table,
	cmd,
	idle_timeout,
	hard_timeout,
	priority,
	buffer_id,
	out_port,
	out_group,
	flags,
	match_type,
	match_length) = struct.unpack_from("!BBHIQQBBHHHIIIH2xHH", msg)

	default = ofpfc_default
	if cmd in (OFPFC_DELETE, OFPFC_DELETE_STRICT):
		default = ofpfc_del_default

	ret = _fixed(default,
		cookie=cookie,
		cookie_mask=cookie_mask,
		table=table,
		priority=priority,
		buffer=buffer_id,
		out_port=out_port,
		out_group=out_group,
		idle_timeout=idle_timeout,
		hard_timeout=hard_timeout,
		flags=flags)

	if match_type == OFPMT_OXM:
		rstr = oxm2str(msg[52:52+match_length-4])
		if len(rstr):
			ret.append(rstr)
	else:
		raise ValueError("match_type {:d} not supported".format(match_type))

	istr = inst2str(msg[48+align8(match_length):hdr_length])
	if len(istr):
		ret.append(istr)

	return ",".join(ret)

def mod2extra(msg):
	(hdr_version, hdr_type, hdr_length, hdr_xid,
	cookie,
	cookie_mask,
	table,
	cmd) = struct.unpack_from("!BBHIQQBB", msg)
	if cmd != OFPFC_ADD:
		return dict(command=cmd, xid=hdr_xid)
	return dict(xid=hdr_xid)

def str2flows(rules, type=OFPT_MULTIPART_REPLY, xid=0):
	if type!=OFPT_MULTIPART_REPLY and not rules:
		rules = [{}]
	
	msgs = []
	capture = b""
	for rule in rules:
		info = dict(rule)
		info.update(str2dict(info.pop("flow", "")))
		
		oxm = info["match"]
		length = 4 + len(oxm)
		match = struct.pack("!HH", OFPMT_OXM, length) + oxm
		match += b"\0" * (align8(length)-length)

		inst = info["inst"]

		body = b""
		if type!=OFPT_MULTIPART_REPLY:
			body = struct.pack("!B3xII4xQQ",
				info.get("table", OFPTT_ALL),
				info.get("out_port", OFPP_ANY),
				info.get("out_group", OFPG_ANY),
				info.get("cookie", 0),
				info.get("cookie_mask", 0)) + match
		else:
			body = struct.pack("!HBxIIHHHH4xQQQ",
				48 + len(match) + len(inst),
				info.get("table", 0),
				info.get("duration_sec", 0),
				info.get("duration_nsec", 0),
				info.get("priority", 0),
				info.get("idle_timeout", 0),
				info.get("hard_timeout", 0),
				info.get("flags", 0),
				info.get("cookie", 0),
				info.get("packet_count", 0),
				info.get("byte_count", 0)) + match + inst
		
		if len(capture) + len(body) > 0xffff - 16:
			flag = OFPMPF_REPLY_MORE
			if type==OFPT_MULTIPART_REQUEST:
				flag = OFPMPF_REQ_MORE
			msgs.append(struct.pack("!BBHIHH4x",
				4, type, 16+len(capture), xid,
				OFPMP_FLOW, flag)+capture)
			capture = body
		else:
			capture += body
	
	msgs.append(struct.pack("!BBHIHH4x",
		4, type, 16+len(capture), xid,
		OFPMP_FLOW, 0)+capture)
	return msgs

def flows2str(msg):
	(hdr_version, hdr_type, hdr_length, hdr_xid,
	mp_type,
	mp_flags) = struct.unpack_from("!BBHIHH4x", msg)
	
	assert mp_type == OFPMP_FLOW, "OFPMP_FLOW required"
	
	rules = []
	body = msg[16:hdr_length]
	if hdr_type != OFPT_MULTIPART_REPLY:
		(table_id,
		out_port,
		out_group,
		cookie,
		cookie_mask,
		match_type,
		match_length) = struct.unpack_from("!B3xII4xQQHH", body)
		defaults = dict(
			table = OFPTT_ALL,
			out_port = OFPP_ANY,
			out_group = OFPG_ANY,
			cookie=0,
			cookie_mask=0)
		ret = _fixed(defaults,
			table=table_id,
			out_port=out_port,
			out_group=out_group,
			cookie=cookie,
			cookie_mask=cookie_mask)
		if match_type == OFPMT_OXM:
			rstr = oxm2str(body[36:36+match_length-4])
			if len(rstr):
				ret.append(rstr)
		else:
			raise ValueError("match_type {:d} not supported".format(match_type))
		rules.append(dict(flow=",".join(ret)))
	else:
		while len(body) >= 56:
			(length,
			table_id,
			duration_sec,
			duration_nsec,
			priority,
			idle_timeout,
			hard_timeout,
			flags,
			cookie,
			packet_count,
			byte_count,
			match_type,
			match_length) = struct.unpack_from("!HBxIIHHHH4xQQQHH", body)

			defaults = dict(
				cookie = 0,
				table = 0,
				priority = 0,
				idle_timeout = 0,
				hard_timeout = 0,
				flags = 0)
			ret = _fixed(defaults,
				cookie=cookie,
				table=table_id,
				priority=priority,
				idle_timeout=idle_timeout,
				hard_timeout=hard_timeout,
				flags=flags)

			if match_type == OFPMT_OXM:
				rstr = oxm2str(body[52:52+match_length-4])
				if len(rstr):
					ret.append(rstr)
			else:
				raise ValueError("match_type {:d} not supported".format(match_type))

			istr = inst2str(body[48+align8(match_length):length])
			if len(istr):
				ret.append(istr)

			rules.append(dict(
				flow=",".join(ret),
				duration_sec=duration_sec,
				duration_nsec=duration_nsec,
				packet_count=packet_count,
				byte_count=byte_count))
			body = body[length:]
	return rules

def flows2extra(msg):
	(hdr_version, hdr_type, hdr_length, hdr_xid,
	mp_type,
	mp_flags) = struct.unpack_from("!BBHIHH4x", msg)
	
	assert mp_type == OFPMP_FLOW, "OFPMP_FLOW required"
	
	ret = dict(xid=hdr_xid)
	if hdr_type != OFPT_MULTIPART_REPLY:
		ret["type"] = hdr_type
	
	if mp_flags != 0:
		ret["flags"] = mp_flags
	
	return ret

def str2group(buckets, command=OFPGC_ADD, group_type=OFPGT_ALL, group_id=0, xid=0):
	body = str2buckets(buckets, group_type=group_type)
	return struct.pack("!BBHIHBxI",
		4, OFPT_GROUP_MOD, 16+len(body), xid,
		command, group_type, group_id) + body

def group2str(msg):
	(hdr_version, hdr_type, hdr_length, hdr_xid,
	command,
	group_type,
	group_id) = struct.unpack_from("!BBHIHBxI", msg)
	assert hdr_type == OFPT_GROUP_MOD
	return buckets2str(msg[16:hdr_length], group_type)

def group2extra(msg):
	(hdr_version, hdr_type, hdr_length, hdr_xid,
	command,
	group_type,
	group_id) = struct.unpack_from("!BBHIHBxI", msg)
	
	assert hdr_type == OFPT_GROUP_MOD
	
	ret = dict(xid=hdr_xid, group_id=group_id)
	if command != OFPGC_ADD:
		ret["command"] = command
	if group_type != OFPGT_ALL:
		ret["group_type"] = group_type
	return ret

def str2groups_desc(groups, type=OFPT_MULTIPART_REPLY, xid=0):
	if type != OFPT_MULTIPART_REPLY:
		return [struct.pack("!BBHIHH4x",
			4, type, 16, xid, OFPMP_GROUP_DESC, 0)]
	
	msgs = []
	capture = b""
	for g in groups:
		bs = str2buckets(g.get("buckets", []),
			group_type=g.get("group_type", OFPGT_ALL))
		body = struct.pack("!HBxI",
			8+len(bs),
			g.get("group_type", OFPGT_ALL),
			g.get("group_id", 0)) + bs
		if len(body)+len(capture) > 0xffff - 16:
			msgs.append(struct.pack("!BBHIHH4x",
				4, type, 16+len(capture), xid,
				OFPMP_GROUP_DESC,
				OFPMPF_REPLY_MORE)+capture)
			capture = body
		else:
			capture += body
	msgs.append(struct.pack("!BBHIHH4x",
		4, type, 16+len(capture), xid,
		OFPMP_GROUP_DESC,
		0)+capture)
	return msgs

def groups_desc2str(msg):
	(hdr_version, hdr_type, hdr_length, hdr_xid,
	mp_type,
	mp_flags) = struct.unpack_from("!BBHIHH4x", msg)
	
	assert mp_type == OFPMP_GROUP_DESC, "OFPMP_GROUP_DESC required"
	
	if hdr_type != OFPT_MULTIPART_REPLY:
		return []
	
	groups = []
	r = msg[16:hdr_length]
	while len(r) >= 8:
		(length, group_type, group_id) = struct.unpack_from("!HBxI", r)
		assert length>=8
		group = dict(
			group_id=group_id,
			buckets=buckets2str(r[8:length], group_type=group_type))
		if group_type != OFPGT_ALL:
			group["group_type"] = group_type
		groups.append(group)
		r = r[length:]
	
	return groups

def groups_desc2extra(msg):
	(hdr_version, hdr_type, hdr_length, hdr_xid,
	mp_type,
	mp_flags) = struct.unpack_from("!BBHIHH4x", msg)
	
	assert mp_type == OFPMP_GROUP_DESC, "OFPMP_GROUP_DESC required"
	
	ret = dict(xid=hdr_xid)
	if hdr_type != OFPT_MULTIPART_REPLY:
		ret["type"] = hdr_type
	return ret



import unittest
import ofpstr.ofp4

class TestRoundTrip(unittest.TestCase):
	actions = (
		"output=controller",
		"output=1",
		"copy_ttl_out",
		"copy_ttl_in",
		"set_mpls_ttl=3",
		"dec_mpls_ttl",
		"push_vlan=0x8100",
		"pop_vlan",
		"push_mpls=0x8847",
		"pop_mpls=0x0800",
		"set_queue=1",
		"group=1",
		"set_nw_ttl=3",
		"dec_nw_ttl",
		"set_vlan_vid=0x10",
		"push_pbb=0x88e7",
		"pop_pbb",
		"cnt_ids(0x1,0x2,0x3)",
		"reg_load(nxm_vlan_tci=0xa/0x0fff)",
		"reg_load2(nxm_vlan_tci=0xa/0x0fff)",
		"reg_move(nxm_eth_dst=nxm_eth_src)",
		"reg_move(nxm_eth_dst[0:4]=nxm_eth_src[4:8])",
		"resubmit(in_port)",
		"resubmit(1)",
		"resubmit_table(in_port,all)",
		"resubmit_table(1,1)",
		"set_tunnel(0x11223344)",
		"set_tunnel64(0x1122334455667788)",
		"pop_queue",
		"note(openflow)",
		"multipath(eth_src,50,hrw,12,0x0,nxm_reg0[0:4])",
		"bundle(eth_src,50,active_backup,nxm_in_port,slaves(1,2,3))",
		"bundle_load(symmetric_l4,60,hrw,nxm_in_port,nxm_reg0[0:16],slaves(2,3))",
		"output_reg(in_port)",
		"output_reg(in_port[0:1])",
		"output_reg2(in_port)",
		"output_reg2(in_port[0:1])",
		"learn(nxm_in_port=0x10,nxm_eth_dst=nxm_eth_src,reg_load(nxm_reg1[16:32]=nxm_in_port))",
		)
	flows = (
		"cookie=0x1",
		"cookie=0x1/0xf",
		"table=2",
		"priority=10",
		"buffer=0x3",
		"idle_timeout=10",
		"hard_timeout=4",
		"@metadata=0x1,@meter=3,@apply,@clear,@write,@goto=5",
		"dot11=1,dot11_frame_ctrl=40/ff,@apply,output=controller",
		"table=1,priority=10,in_port=1,@apply,output=controller,@goto=2",
		"priority=20,idle_timeout=30,dot11=1,dot11_frame_ctrl=00/0f,dot11_addr2=01:23:45:67:89:01,@apply,output=controller",
		"cookie=0x1/0xf,priority=4,buffer=0x1,idle_timeout=300,hard_timeout=300,vlan_vid=0x1,@apply,set_vlan_vid=0x2,output=3,@goto=3",
		)
	del_flows = (
		"out_port=1",
		"out_port=controller",
		"out_group=1",
	)
	aliased_flows = [
		("@apply,set_nxm_tun_ipv4_src(192.168.0.1),set_nxm_tun_ipv4_dst(192.168.0.2)",
			"@apply,set_nxm_tun_ipv4_src=192.168.0.1,set_nxm_tun_ipv4_dst=192.168.0.2"),
		]
	def test_action(self):
		for rule in self.actions:
			msg, length = ofpstr.ofp4.str2act(rule)
			assert length == len(rule), "{:s} captured {:d}".format(rule, length)
			assert rule == ofpstr.ofp4.act2str(msg), "{0} != {1}".format(rule, ofpstr.ofp4.act2str(msg))
	
	def test_mod(self):
		for flow in self.flows:
			ret = ofpstr.ofp4.mod2str(ofpstr.ofp4.str2mod(flow))
			assert ret == flow, "{:s}!={:s}".format(ret, flow)
		for flow in self.del_flows:
			OFPFC_DELETE = 3
			ret = ofpstr.ofp4.mod2str(ofpstr.ofp4.str2mod(flow, command=OFPFC_DELETE))
			assert ret == flow, "{:s}!={:s}".format(ret, flow)
		for aliased, flow in self.aliased_flows:
			ret = ofpstr.ofp4.mod2str(ofpstr.ofp4.str2mod(aliased))
			assert ret == flow, ret

class TestFlowStat(unittest.TestCase):
	def test_empty_req(self):
		OFPT_MULTIPART_REQUEST = 18
		flows = [{"flow":""}]
		msgs = ofpstr.ofp4.str2flows(flows, type=OFPT_MULTIPART_REQUEST)
		assert len(msgs) == 1
		rules = ofpstr.ofp4.flows2str(msgs[0])
		assert rules
		for rule in rules:
			assert not rule["flow"], rule["flow"]

	def test_empty_res(self):
		flows = []
		msgs = ofpstr.ofp4.str2flows(flows)
		assert len(msgs) == 1
		rules = ofpstr.ofp4.flows2str(msgs[0])
		assert not rules

	def test_single_res(self):
		flows = [{"flow":"send_flow_rem,in_port=1", "byte_count":20}]
		msgs = ofpstr.ofp4.str2flows(flows)
		assert len(msgs) == 1
		rules = ofpstr.ofp4.flows2str(msgs[0])
		for r in rules:
			assert r["flow"] == "send_flow_rem,in_port=1", r["flow"]
			assert r["byte_count"] == 20

class TestGroup(unittest.TestCase):
	def test_empty(self):
		assert not ofpstr.ofp4.group2str(ofpstr.ofp4.str2group([]))

	def test_two(self):
		msg = ofpstr.ofp4.str2group([{"actions":"output=0"}, {"actions":"output=1"}])
		rules = ofpstr.ofp4.group2str(msg)
		assert rules[0]["actions"] == "output=0"
		assert rules[1]["actions"] == "output=1"

class TestGroupsDesc(unittest.TestCase):
	def test_empty(self):
		msgs = ofpstr.ofp4.str2groups_desc([])
		assert len(msgs) == 1
		assert not ofpstr.ofp4.groups_desc2str(msgs[0])

	def test_two(self):
		msgs = ofpstr.ofp4.str2groups_desc([
			dict(group_id=1, buckets=[{"actions":"output=0"}, {"actions":"output=1"}]),
			dict(group_id=2, buckets=[{"actions":"output=2"}, {"actions":"output=3"}]),
			])
		assert len(msgs) == 1
		rules = ofpstr.ofp4.groups_desc2str(msgs[0])
		assert len(rules) == 2
		assert rules[0]["buckets"][0]["actions"] == "output=0"
		assert rules[0]["buckets"][1]["actions"] == "output=1"
		assert rules[1]["buckets"][0]["actions"] == "output=2"
		assert rules[1]["buckets"][1]["actions"] == "output=3"


if __name__ == "__main__":
	unittest.main()



import re
import string

try:
	L0 = long(0)
except:
	L0 = 0
	long = int

ofpp = {
	0xffffff00: "max",
	0xfffffff8: "in_port",
	0xfffffff9: "table",
	0xfffffffa: "normal",
	0xfffffffb: "flood",
	0xfffffffc: "all",
	0xfffffffd: "controller",
	0xfffffffe: "local",
	0xffffffff: "any"
	}

def is_delimiter(c):
	return c in "," or c in string.whitespace

pars = dict(("()", "{}", "[]", "<>"))
def scan_paren(end):
	def scan(reader):
		'''
		reader is iterator that yields a char
		@returns payload, right-parenthesis
		'''
		payload = ""
		for c in reader:
			if c == end:
				return payload, c
			else:
				payload += c
				if c in pars:
					p,r = scan_paren(pars[c])(reader)
					payload += p
					if r:
						payload += r
		return payload, ""
	return scan

def get_token(unparsed):
	head = "" # leading non-token string
	m = re.match("^(\s*=\s*|[\s,]+)", unparsed)
	if m:
		head = m.group(1)
	
	body = ""
	reader = iter(unparsed[len(head):])
	for c in reader:
		if is_delimiter(c) or c == "=":
			break
		body += c
		if c in pars:
			p,r = scan_paren(pars[c])(reader)
			body += p + r
	
	return head, body, unparsed[len(head)+len(body):]

def split(unparsed):
	ret = []
	tok = ""
	reader = iter(unparsed)
	for c in reader:
		if is_delimiter(c):
			if tok:
				ret.append(tok)
			tok = ""
		elif c in pars:
			p, r = scan_paren(pars[c])(reader)
			tok += p+r
		else:
			tok += c
	if tok:
		ret.append(tok)
	
	return ret

def parse_func(nojunk):
	'''
	parses argument is function-style string or not.
	This checks argument is cleanly closed with parenthesis
	'''
	name = ""
	reader = iter(nojunk)
	for c in reader:
		if c in pars:
			p, r = scan_paren(pars[c])(reader)
			assert len(nojunk) == len(name)+1+len(p)+len(r)
			# cleanly closed only if len(r)!=0, should we raise warn here?
			return name, p
		else:
			name += c
	
	return name, None

def longest(s, char_set):
	'''returns the maximum continuous length of string, which is made from char_set.'''
	i = 0
	for c in s:
		if c in char_set:
			i += 1
		else:
			break
	return i

def parseInt(unparsed):
	if unparsed.startswith("-"):
		l = 1 + longest(unparsed[1:], "0123456789")
		return -long(unparsed[1:l]), l
	
	if unparsed.startswith("0x") or unparsed.startswith("0X"):
		l = 2 + longest(unparsed[2:], "0123456789abcdefABCDEF")
		return long(unparsed[2:l], 16), l
	elif unparsed.startswith("0") and len(unparsed)>2 and unparsed[1] in "01234567":
		l = 1 + longest(unparsed[1:], "01234567")
		return long(unparsed[1:l], 8), l
	else:
		l = longest(unparsed, "0123456789")
		return long(unparsed[:l]), l

def parseFloat(unparsed):
	neg = False
	if unparsed.startswith("-"):
		neg = True
		unparsed = unparsed[1:]
	
	l = longest(unparsed, "0123456789")
	ret = long(unparsed[:l])
	if unparsed[l:].startswith("."):
		l2 = longest(unparsed[l+1:], "0123456789")
		ret += float(long(unparsed[l+1:l+l2+1]))/(10**l2)
		l += l2+1
	
	return ret, l

try:
	unichr(0)
	int2bytes = lambda seq: b"".join(map(chr, seq))
except:
	int2bytes = bytes



import unittest
import ofpstr.util

class TestToken(unittest.TestCase):
	rules = (
		("a,b,c", ("", "a", ",b,c")), # comma separated list
		(",b,c", (",", "b", ",c")),
		(",c", (",", "c", "")),
		("a b c", ("", "a", " b c")), # whitespace separated list
		(" b c", (" ", "b", " c")),
		(" c", (" ", "c", "")),
		("a = b, c = d", ("", "a", " = b, c = d")), # mixture
		(" = b, c = d", (" = ", "b", ", c = d")),
		("learn(hoge),learn(geho)", ("", "learn(hoge)", ",learn(geho)")), # parenthesis
	)
	def test_expected(self):
		for input,output in self.rules:
			processed = ofpstr.util.get_token(input)
			assert output == processed, "{0}=>{1}".format(input, processed)

if __name__ == "__main__":
	unittest.main()



import struct
import re
import socket
import binascii

from .util import get_token, parseInt, parseFloat, ofpp, int2bytes

try:
	L0 = long(0)
except:
	L0 = 0
	long = int

try:
	strtypes = (str, unicode)
except:
	strtypes = (str,)

OFPXMC_NXM_0 = 0x0000
OFPXMC_NXM_1 = 0x0001
OFPXMC_OPENFLOW_BASIC = 0x8000
OFPXMC_PACKET_REGS = 0x8001
OFPXMC_EXPERIMENTER = 0xFFFF

ofb_names = '''in_port
	in_phy_port
	metadata
	eth_dst
	eth_src
	eth_type
	vlan_vid
	vlan_pcp
	ip_dscp
	ip_ecn
	ip_proto
	ipv4_src
	ipv4_dst
	tcp_src
	tcp_dst
	udp_src
	udp_dst
	sctp_src
	sctp_dst
	icmpv4_type
	icmpv4_code
	arp_op
	arp_spa
	arp_tpa
	arp_sha
	arp_tha
	ipv6_src
	ipv6_dst
	ipv6_flabel
	icmpv6_type
	icmpv6_code
	ipv6_nd_target
	ipv6_nd_sll
	ipv6_nd_tll
	mpls_label
	mpls_tc
	mpls_bos
	pbb_isid
	tunnel_id
	ipv6_exthdr
	_
	pbb_uca
	tcp_flags
	actset_output
	packet_type'''.split()
for (i,n) in enumerate(ofb_names):
	if n != "_":
		globals()["OXM_OF_{:s}".format(n.upper())] = i

class nxm_of(int):
	def __hash__(self):
		return hash((0, int(self)))

nxm0_names = '''in_port
	eth_dst
	eth_src
	eth_type
	vlan_tci
	ip_tos
	ip_proto
	ip_src
	ip_dst
	tcp_src
	tcp_dst
	udp_src
	udp_dst
	icmp_type
	icmp_code
	arp_op
	arp_spa
	arp_tpa'''.split()
for (i, n) in enumerate(nxm0_names):
	if n != "_":
		globals()["NXM_OF_{:s}".format(n.upper())] = nxm_of(i)

class nxm_nx(int):
	def __hash__(self):
		return hash((1, int(self)))

nxm1_names = '''reg0
	reg1
	reg2
	reg3
	reg4
	reg5
	reg6
	reg7
	_ _ _ _ _ _ _ _
	tun_id
	arp_sha
	arp_tha
	ipv6_src
	ipv6_dst
	icmpv6_type
	icmpv6_code
	nd_target
	nd_sll
	nd_tll
	ip_frag
	ipv6_label
	ip_ecn
	ip_ttl
	_
	tun_ipv4_src
	tun_ipv4_dst
	pkt_mark
	tcp_flags
	dp_hash
	recirc_id
	conj_id
	tun_gbp_id
	tun_gbp_flags'''.split()
for (i, n) in enumerate(nxm1_names):
	if n != "_":
		globals()["NXM_NX_{:s}".format(n.upper())] = nxm_nx(i)
assert NXM_NX_TUN_GBP_FLAGS==39

STRATOS_EXPERIMENTER_ID = 0xFF00E04D

STRATOS_OXM_FIELD_BASIC = 0
STRATOS_OXM_FIELD_RADIOTAP = 1

class stratos(int):
	def __hash__(self):
		return hash((OFPXMC_EXPERIMENTER, STRATOS_EXPERIMENTER_ID, int(self)))

stratos_names = '''_ _
	dot11
	dot11_frame_ctrl
	dot11_addr1
	dot11_addr2
	dot11_addr3
	dot11_addr4
	dot11_ssid
	dot11_action_category
	dot11_public_action
	dot11_tag
	dot11_tag_vendor
	_ _ _
	tsft
	flags
	rate
	channel
	fhss
	dbm_antsignal
	dbm_antnoise
	lock_quality
	tx_attenuation
	db_tx_attenuation
	dbm_tx_power
	antenna
	db_antsignal
	db_antnoise
	rx_flags
	tx_flags
	rts_retries
	data_retries
	_
	mcs
	ampdu_status
	vht'''.split()
for (i,n) in enumerate(stratos_names):
	if n != "_":
		if i < 16:
			vname = "STROXM_BASIC_{:s}".format(n.upper())
		else:
			vname = "STROXM_RADIOTAP_{:s}".format(n.upper())
		globals()[vname] = stratos(i)
assert STROXM_RADIOTAP_VHT == 37


_bin2str = {}
_str2bin = {}


def header(field, value_length, mask):
	length = value_length
	field_HM = field<<1
	if mask:
		field_HM += 1
		length *= 2
	if isinstance(field, stratos):
		length += 4
	
	if length > 0xff:
		length = 0xff
	
	if isinstance(field, stratos):
		return struct.pack("!HBBI", OFPXMC_EXPERIMENTER, field_HM, length, STRATOS_EXPERIMENTER_ID)
	elif isinstance(field, nxm_of):
		return struct.pack("!HBB", OFPXMC_NXM_0, field_HM, length)
	elif isinstance(field, nxm_nx):
		return struct.pack("!HBB", OFPXMC_NXM_1, field_HM, length)
	else:
		return struct.pack("!HBB", OFPXMC_OPENFLOW_BASIC, field_HM, length)


def uint_bin2str(fmt):
	def bin2str(payload, has_mask):
		if payload is None:
			return fmt.split("=")[0]
		
		l = len(payload)
		if has_mask:
			value = L0
			mask = L0
			
			split = l//2
			for v in struct.unpack_from("!{:d}B".format(split), payload):
				value = (value << 8) + v
			for v in struct.unpack_from("!{:d}B".format(len(payload)-split), payload, split):
				mask = (mask << 8) + v
			
			return fmt.format(value, mask)
		else:
			value = L0
			for c in struct.unpack("!{:d}B".format(len(payload)), payload):
				value = (value << 8) + c
			
			return fmt.split("/", 2)[0].format(value)
	
	return bin2str


def uint_str2bin(field, size):
	def str2bin(unparsed):
		rlen = 0
		has_mask = False
		payload = b""
		if isinstance(unparsed, strtypes):
			num,rlen = parseInt(unparsed)
			payload += int2bytes([(num>>(8*s))&0xff for s in reversed(range(size))])
			
			if unparsed[rlen:].startswith("/"):
				has_mask = True
				num,e = parseInt(unparsed[rlen+1:])
				payload += int2bytes([(num>>(8*s))&0xff for s in reversed(range(size))])
				rlen += 1 + e
		elif unparsed:
			has_mask = True
		
		return header(field, size, has_mask)+payload, rlen
	
	return str2bin


def port_bin2str(name, fmt="!I"):
	size = struct.calcsize(fmt)
	mask = (1<<size*8) - 1
	def bin2str(payload, has_mask):
		if payload is None:
			return name
		
		assert not has_mask, "{:s} does not take mask".format(name)
		assert len(payload) == size, repr(payload)
		num = struct.unpack_from(fmt, payload)[0]
		for p,port in ofpp.items():
			if p&mask == num:
				return "{:s}={:s}".format(name, port)
		return "{:s}={:d}".format(name, num)
	
	return bin2str


def port_str2bin(field, fmt="!I"):
	size = struct.calcsize(fmt)
	mask = (1<<size*8) - 1
	def str2bin(unparsed):
		rlen = 0
		payload = b""
		if isinstance(unparsed, strtypes):
			for (v,name) in ofpp.items():
				for nm in (name.lower(), name.upper()):
					if unparsed.startswith(nm):
						rlen = len(name)
						payload = struct.pack(fmt, v&mask)
						break
			if not payload:
				num, rlen = parseInt(unparsed)
				payload = struct.pack(fmt, num)
		
		return header(field, size, False)+payload, rlen
	
	return str2bin


def mac_bin2str(name):
	def bin2str(payload, has_mask):
		if payload is None:
			return name
		
		value = ":".join(map("{:02x}".format, struct.unpack("!6B", payload[:6])))
		if has_mask:
			assert len(payload) == 12
			return "{:s}={:s}/{:s}".format(name, value, ":".join(map("{:02x}".format, struct.unpack("!6B", payload[6:]))))
		else:
			assert len(payload) == 6
			return "{:s}={:s}".format(name, value)
	
	return bin2str


def mac_str2bin(field):
	def scan(txt):
		for r,l in (("([0-9A-Fa-f]{12})", 12),
				(".".join(["([0-9A-Fa-f]{4})"]*3), 14),
				(":".join(["([0-9A-Fa-f]{2})"]*6), 17),
				("-".join(["([0-9A-Fa-f]{2})"]*6), 17)):
			m = re.match("^"+r, txt)
			if m:
				return "".join(m.groups()), l
		
		raise ValueError("mac format error {:s}".format(txt))
	
	def str2bin(unparsed):
		rlen = 0
		payload = b""
		has_mask = False
		if isinstance(unparsed, strtypes):
			value, rlen = scan(unparsed)
			if unparsed[rlen:].startswith("/"):
				mask, l = scan(unparsed[rlen+1:])
				rlen += 1+l
				payload = binascii.a2b_hex(value+mask)
				has_mask = True
			else:
				payload = binascii.a2b_hex(value)
		elif unparsed:
			has_mask = True
		
		return header(field, 6, has_mask)+payload, rlen
	
	return str2bin


def ipv4_bin2str(name):
	def bin2str(payload, has_mask):
		if payload is None:
			return name
		
		value = socket.inet_ntoa(payload[:4])
		if has_mask:
			assert len(payload) == 8
			return "{:s}={:s}/{:s}".format(name, value,  socket.inet_ntoa(payload[4:]))
		else:
			assert len(payload) == 4
			return "{:s}={:s}".format(name, value)
	
	return bin2str


def ipv4_str2bin(field):
	def str2bin(unparsed):
		rlen = 0
		payload = b""
		has_mask = False
		if isinstance(unparsed, strtypes):
			h, txt, unparsed = get_token(unparsed)
			vm = txt.split("/", 1)
			payload = socket.inet_aton(vm[0])
			if len(vm) > 1:
				has_mask = True
				try:
					mask = socket.inet_aton(vm[1])
				except ValueError:
					mask_len = int(vm[1])
					mask = struct.pack("!I", 0xFFFFFFFF<<(32-mask_len))
				
				payload += mask
			rlen = len(h)+len(txt)
		elif unparsed:
			has_mask = True
		
		return header(field, 4, has_mask)+payload, rlen
	
	return str2bin


def ipv6_bin2str(name):
	def bin2str(payload, has_mask):
		if payload is None:
			return name
		
		value = socket.inet_ntop(socket.AF_INET6, payload[:16])
		if has_mask:
			assert len(payload) == 32
			return "{:s}={:s}/{:s}".format(name, value,  socket.inet_ntop(socket.AF_INET6, payload[16:]))
		else:
			assert len(payload) == 16
			return "{:s}={:s}".format(name, value)
	
	return bin2str


def ipv6_str2bin(field):
	def str2bin(unparsed):
		rlen = 0
		payload = b""
		has_mask = False
		if isinstance(unparsed, strtypes):
			h,txt,unparsed = get_token(unparsed)
			rlen = len(h)+len(txt)
			vm = txt.split("/", 1)
			payload = socket.inet_pton(socket.AF_INET6, vm[0])
			if len(vm) > 1:
				has_mask = True
				try:
					mask = socket.inet_pton(socket.AF_INET6, vm[1])
				except:
					mask_len = int(vm[1])
					mask_int = 0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF<<(128-int(mask_len))
					mask = struct.pack("!QQ", mask_int>>64, mask_int&0xFFFFFFFFFFFFFFFF)
				
				payload += mask
		elif unparsed:
			has_mask = True
		
		return header(field, 16, has_mask)+payload, rlen
	
	return str2bin


def pkt_bin2str(payload, has_mask):
	if payload is None:
		return "packet_type"
	
	assert not has_mask
	assert len(payload) == 4
	return "packet_type={:#x}:{:#x}".format(*struct.unpack("!HH", payload))


def pkt_str2bin(unparsed):
	rlen = 0
	payload = b""
	if isinstance(unparsed, strtypes):
		ns, l = parseInt(unparsed)
		assert unparsed[l] == ":"
		ns_type, l2 = parseInt(unparsed[l+1:])
		
		rlen = l+l2+1
		payload = struct.pack("!HH", ns, ns_type)
	
	return header(OXM_OF_PACKET_TYPE, 4, False)+payload, rlen


def hex_bin2str(name):
	def bin2str(payload, has_mask):
		if payload is None:
			return name
		
		if has_mask:
			split = len(payload)//2
			return "{:s}={:s}/{:s}".format(name,
				binascii.b2a_hex(payload[:split]).decode("UTF-8"),
				binascii.b2a_hex(payload[split:]).decode("UTF-8"))
		else:
			return "{:s}={:s}".format(name,
				binascii.b2a_hex(payload).decode("UTF-8"))
	
	return bin2str


def hex_str2bin(field):
	def str2bin(unparsed):
		rlen = 0
		payload = b""
		has_mask = False
		length = 0xff
		if isinstance(unparsed, strtypes):
			h, txt, unparsed = get_token(unparsed)
			rlen = len(h)+len(txt)
			vm = txt.split("/", 1)
			payload = binascii.a2b_hex(vm[0])
			length = len(payload)
			if len(vm) > 1:
				has_mask = True
				mask = binascii.a2b_hex(vm[1])
				pad = b"\0" * abs(len(mask)-len(payload))
				if len(payload) < len(mask):
					length = len(mask)
					payload = payload+pad+mask
				else:
					payload = payload+mask+pad
		elif unparsed:
			has_mask = True
		
		return header(field, length, has_mask)+payload, rlen
	
	return str2bin


def ssid_bin2str(payload, has_mask):
	if payload is None:
		return "dot11_ssid"
	
	if has_mask:
		split = len(payload)//2
		name = bytearray(payload[:split])
		while name[-1] == 0:
			name = name[:-1]
		name = bytes(name)
		
		return "dot11_ssid={:s}/{:s}".format(
			binascii.b2a_qp(name).decode("UTF-8"),
			binascii.b2a_hex(payload[split:]).decode("UTF-8"))
	else:
		return "dot11_ssid={:s}".format(binascii.b2a_qp(payload).decode("UTF-8"))


def ssid_str2bin(unparsed):
	rlen = 0
	payload = b""
	has_mask = False
	length = 0xff
	if isinstance(unparsed, strtypes):
		h, txt, unparsed = get_token(unparsed)
		rlen = len(h) + len(txt)
		vm = txt.split("/")
		payload = binascii.a2b_qp(vm[0])
		length = len(payload)
		if len(vm) > 1:
			has_mask = True
			mask = binascii.a2b_hex(vm[1])
			pad = b"\0" * abs(len(payload)-len(mask))
			if len(payload) < len(mask):
				length = len(mask)
				payload += pad + mask
			else:
				payload += mask + pad
	elif unparsed:
		has_mask = True
	
	return header(STROXM_BASIC_DOT11_SSID, length, has_mask)+payload, rlen


def le_bin2str(fmt):
	def bin2str(payload, has_mask):
		if payload is None:
			return fmt.split("=")[0]
		
		l = len(payload)
		if has_mask:
			value = L0
			mask = L0
			
			split = l//2
			for i,v in enumerate(struct.unpack_from("!{:d}B".format(split), payload)):
				value += v<<(8*i)
			for i,v in enumerate(struct.unpack_from("!{:d}B".format(len(payload)-split), payload, split)):
				mask += v<<(8*i)
			
			return fmt.format(value, mask)
		else:
			value = L0
			for i,c in enumerate(struct.unpack("!{:d}B".format(len(payload)), payload)):
				value += c<<(8*i)
			
			return fmt.split("/", 2)[0].format(value)
	
	return bin2str


def le_str2bin(field, size):
	def str2bin(unparsed):
		rlen = 0
		payload = b""
		has_mask = False
		if isinstance(unparsed, strtypes):
			num,rlen = parseInt(unparsed)
			payload += int2bytes([0xFF & (num>>(8*s)) for s in range(size)])
			
			if unparsed[rlen:].startswith("/"):
				has_mask = True
				num,e = parseInt(unparsed[rlen+1:])
				rlen += 1 + e
				for s in range(size):
					payload += chr(0xFF & (num>>(8*s)))
		elif unparsed:
			has_mask = True
		
		return header(field, size, has_mask)+payload, rlen
	
	return str2bin


def rate_bin2str(payload, has_mask):
	if payload is None:
		return "radiotap_rate"
	
	assert not has_mask
	rate = struct.unpack_from("!B", payload)[0]
	if rate < 2:
		return "radiotap_rate={:.1f}K".format(rate*500.0)
	else:
		return "radiotap_rate={:.1f}M".format(float(rate)/2.0)


def rate_str2bin(unparsed):
	rlen = 0
	payload = b""
	has_mask = False
	if isinstance(unparsed, strtypes):
		num, rlen = parseFloat(unparsed)
		if unparsed[rlen] == "K":
			rate = num / 500
			rlen += 1
		elif unparsed[rlen] == "M":
			rate = num * 2
			rlen += 1
		else:
			raise ValueError("rate unit error")
		
		payload = struct.pack("!B", int(rate))
	
	return header(STROXM_RADIOTAP_RATE, 1, False)+payload, rlen


def ch_bin2str(payload, has_mask):
	if payload is None:
		return "radiotap_channel"
	
	if has_mask:
		assert len(payload) == 8
		v = struct.unpack("!HHHH", payload)
		if v[2] == 0:
			return "radiotap_channel={:d}:{:#06x}/:{:#06x}".format(
				v[0], v[1], v[3])
		else:
			return "radiotap_channel={:d}:{:#06x}/{:#06x}:{:#06x}".format(*v)
	else:
		assert len(payload) == 4
		v = struct.unpack("!HH", payload)
		return "radiotap_channel={:d}:{:#06x}".format(*v)


def ch_str2bin(unparsed):
	rlen = 0
	payload = b""
	has_mask = False
	
	if isinstance(unparsed, strtypes):
		v1,l = parseInt(unparsed)
		assert unparsed[l] == ":"
		v2,l2 = parseInt(unparsed[l+1:])
		payload = struct.pack("!HH", v1, v2)
		rlen = l+1+l2
		
		if unparsed[rlen:].startswith("/"):
			has_mask = True
			if unparsed[rlen+1:].startswith(":"):
				m1 = l3 = 0
			else:
				m1,l3 = parseInt(unparsed[rlen+1:])
			
			if unparsed[rlen+1+l3:].startswith(":"):
				m2,l4 = parseInt(unparsed[rlen+1+l3+1:])
				rlen += 1 + l3 + 1 + l4
			else:
				m2 = 0
				rlen += 1 + l3
			
			payload += struct.pack("!HH", m1, m2)
	elif unparsed:
		has_mask = True
	
	return header(STROXM_RADIOTAP_CHANNEL, 4, has_mask)+payload, rlen


def comp_bin2str(name, packs, fmts):
	def collect(data):
		ret = []
		for n,fmt in zip(struct.unpack_from(packs, data), fmts):
			if n == 0:
				ret.append("")
			else:
				ret.append(fmt.format(n))
		return ":".join(ret)
	
	def bin2str(payload, has_mask):
		if payload is None:
			return name
		
		ret = "{:s}={:s}".format(name, collect(payload))
		if has_mask:
			ret += "/"
			ret += collect(payload[struct.calcsize(packs):])
		
		return ret

	return bin2str


def comp_str2bin(field, packs):
	def collect(txt):
		ps = []
		for t in txt.split(":"):
			v = l = 0
			if len(t):
				v, l = parseInt(t)
				assert len(t) == l
			ps.append(v)
		
		return struct.pack(packs, *ps)
	
	def str2bin(unparsed):
		rlen = 0
		payload = b""
		has_mask = False
		
		if isinstance(unparsed, strtypes):
			h,txt,unparsed = get_token(unparsed)
			rlen = len(h) + len(txt)
			vm = txt.split("/", 1)
			payload = collect(vm[0])
			if len(vm) > 1:
				has_mask = True
				payload += collect(vm[1])
		elif unparsed:
			has_mask = True
		
		return header(field, struct.calcsize(packs), has_mask)+payload, rlen
	
	return str2bin


def vht_bin2str(payload, has_mask):
	name = "radiotap_vht"
	if payload is None:
		return name
	
	packs = "<HBB4BBBH"
	def collect(data):
		ns = struct.unpack_from(packs, data)
		
		ret = ""
		args = []
		if ns[0]>0:
			ret += "{:#06x}".format(ns[0])
		ret += ":"
		if ns[1]>0:
			ret += "{:#04x}".format(ns[1])
		ret += ":{:d}:".format(ns[2])
		if sum(ns[3:7])!=0:
			ret += "{:02x}{:02x}{:02x}{:02x}".format(*ns[3:7])
		ret += ":"
		if ns[7]>0:
			ret += "{:#04x}".format(ns[7])
		ret += ":{:d}:{:#06x}".format(ns[8], ns[9])
		return ret
	
	ret = "{:s}={:s}".format(name, collect(payload))
	if has_mask:
		ret += "/"
		ret += collect(payload[struct.calcsize(packs):])
	
	return ret


def vht_str2bin(unparsed):
	packs = "<HBB4BBBH"
	def collect(txt):
		ps = []
		for i, t in enumerate(txt.split(":")):
			if i == 3:
				ps.extend(bytearray(binascii.a2b_hex(t)))
			else:
				v = l = 0
				if len(t):
					v, l = parseInt(t)
				ps.append(v)
		
		return struct.pack(packs, *ps)
	
	rlen = 0
	payload = b""
	has_mask = False
	if isinstance(unparsed, strtypes):
		h,txt,unparsed = get_token(unparsed)
		rlen = len(h) + len(txt)
		vm = txt.split("/", 1)
		payload = collect(vm[0])
		if len(vm) > 1:
			has_mask = True
			payload += collect(vm[1])
	elif unparsed:
		has_mask = True
	
	return header(STROXM_RADIOTAP_VHT, struct.calcsize(packs), has_mask)+payload, rlen


def s8_bin2str(name):
	def bin2str(payload, has_mask):
		if payload is None:
			return name
		
		assert not has_mask
		return "{:s}={:d}".format(name, struct.unpack("<b", payload)[0])
	
	return bin2str


def s8_str2bin(field):
	def str2bin(unparsed):
		rlen = 0
		payload = b""
		if isinstance(unparsed, strtypes):
			num,rlen = parseInt(unparsed)
			payload = struct.pack("<b", num)
		
		return header(field, 1, False)+payload, rlen
	
	return str2bin


_bin2str[OXM_OF_IN_PORT] = port_bin2str("in_port")
_str2bin["in_port"] = port_str2bin(OXM_OF_IN_PORT)

_bin2str[OXM_OF_IN_PHY_PORT] = port_bin2str("in_phy_port")
_str2bin["in_phy_port"] = port_str2bin(OXM_OF_IN_PHY_PORT)

_bin2str[OXM_OF_METADATA] = uint_bin2str("metadata={:#x}/{:#x}")
_str2bin["metadata"] = uint_str2bin(OXM_OF_METADATA, 8)

_bin2str[OXM_OF_ETH_DST] = mac_bin2str("eth_dst")
_str2bin["eth_dst"] = mac_str2bin(OXM_OF_ETH_DST)

_bin2str[OXM_OF_ETH_SRC] = mac_bin2str("eth_src")
_str2bin["eth_src"] = mac_str2bin(OXM_OF_ETH_SRC)

_bin2str[OXM_OF_ETH_TYPE] = uint_bin2str("eth_type={:#04x}")
_str2bin["eth_type"] = uint_str2bin(OXM_OF_ETH_TYPE, 2)

_bin2str[OXM_OF_VLAN_VID] = uint_bin2str("vlan_vid={:#x}/{:#x}")
_str2bin["vlan_vid"] = uint_str2bin(OXM_OF_VLAN_VID, 2)

_bin2str[OXM_OF_VLAN_PCP] = uint_bin2str("vlan_pcp={:d}")
_str2bin["vlan_pcp"] = uint_str2bin(OXM_OF_VLAN_PCP, 1)

_bin2str[OXM_OF_IP_DSCP] = uint_bin2str("ip_dscp={:#x}")
_str2bin["ip_dscp"] = uint_str2bin(OXM_OF_IP_DSCP, 1)

_bin2str[OXM_OF_IP_ECN] = uint_bin2str("ip_ecn={:#x}")
_str2bin["ip_ecn"] = uint_str2bin(OXM_OF_IP_ECN, 1)

_bin2str[OXM_OF_IP_PROTO] = uint_bin2str("ip_proto={:d}")
_str2bin["ip_proto"] = uint_str2bin(OXM_OF_IP_PROTO, 1)

_bin2str[OXM_OF_IPV4_SRC] = ipv4_bin2str("ipv4_src")
_str2bin["ipv4_src"] = ipv4_str2bin(OXM_OF_IPV4_SRC)

_bin2str[OXM_OF_IPV4_DST] = ipv4_bin2str("ipv4_dst")
_str2bin["ipv4_dst"] = ipv4_str2bin(OXM_OF_IPV4_DST)

_bin2str[OXM_OF_TCP_SRC] = uint_bin2str("tcp_src={:d}")
_str2bin["tcp_src"] = uint_str2bin(OXM_OF_TCP_SRC, 2)

_bin2str[OXM_OF_TCP_DST] = uint_bin2str("tcp_dst={:d}")
_str2bin["tcp_dst"] = uint_str2bin(OXM_OF_TCP_DST, 2)

_bin2str[OXM_OF_UDP_SRC] = uint_bin2str("udp_src={:d}")
_str2bin["udp_src"] = uint_str2bin(OXM_OF_UDP_SRC, 2)

_bin2str[OXM_OF_UDP_DST] = uint_bin2str("udp_dst={:d}")
_str2bin["udp_dst"] = uint_str2bin(OXM_OF_UDP_DST, 2)

_bin2str[OXM_OF_SCTP_SRC] = uint_bin2str("sctp_src={:d}")
_str2bin["sctp_src"] = uint_str2bin(OXM_OF_SCTP_SRC, 2)

_bin2str[OXM_OF_SCTP_DST] = uint_bin2str("sctp_dst={:d}")
_str2bin["sctp_dst"] = uint_str2bin(OXM_OF_SCTP_DST, 2)

_bin2str[OXM_OF_ICMPV4_TYPE] = uint_bin2str("icmpv4_type={:d}")
_str2bin["icmpv4_type"] = uint_str2bin(OXM_OF_ICMPV4_TYPE, 1)

_bin2str[OXM_OF_ICMPV4_CODE] = uint_bin2str("icmpv4_code={:d}")
_str2bin["icmpv4_code"] = uint_str2bin(OXM_OF_ICMPV4_CODE, 1)

_bin2str[OXM_OF_ARP_OP] = uint_bin2str("arp_op={:#d}")
_str2bin["arp_op"] = uint_str2bin(OXM_OF_ARP_OP, 2)

_bin2str[OXM_OF_ARP_SPA] = ipv4_bin2str("arp_spa")
_str2bin["arp_spa"] = ipv4_str2bin(OXM_OF_ARP_SPA)

_bin2str[OXM_OF_ARP_TPA] = ipv4_bin2str("arp_tpa")
_str2bin["arp_tpa"] = ipv4_str2bin(OXM_OF_ARP_TPA)

_bin2str[OXM_OF_ARP_SHA] = mac_bin2str("arp_sha")
_str2bin["arp_sha"] = mac_str2bin(OXM_OF_ARP_SHA)

_bin2str[OXM_OF_ARP_THA] = mac_bin2str("arp_tha")
_str2bin["arp_tha"] = mac_str2bin(OXM_OF_ARP_THA)

_bin2str[OXM_OF_IPV6_SRC] = ipv6_bin2str("ipv6_src")
_str2bin["ipv6_src"] = ipv6_str2bin(OXM_OF_IPV6_SRC)

_bin2str[OXM_OF_IPV6_DST] = ipv6_bin2str("ipv6_dst")
_str2bin["ipv6_dst"] = ipv6_str2bin(OXM_OF_IPV6_DST)

_bin2str[OXM_OF_IPV6_FLABEL] = uint_bin2str("ipv6_flabel={:#x}/{:#x}")
_str2bin["ipv6_flabel"] = uint_str2bin(OXM_OF_IPV6_FLABEL, 4)

_bin2str[OXM_OF_ICMPV6_TYPE] = uint_bin2str("icmpv6_type={:d}")
_str2bin["icmpv6_type"] = uint_str2bin(OXM_OF_ICMPV6_TYPE, 1)

_bin2str[OXM_OF_ICMPV6_CODE] = uint_bin2str("icmpv6_code={:d}")
_str2bin["icmpv6_code"] = uint_str2bin(OXM_OF_ICMPV6_CODE, 1)

_bin2str[OXM_OF_IPV6_ND_TARGET] = ipv6_bin2str("ipv6_nd_target")
_str2bin["ipv6_nd_target"] = ipv6_str2bin(OXM_OF_IPV6_ND_TARGET)

_bin2str[OXM_OF_IPV6_ND_SLL] = mac_bin2str("ipv6_nd_sll")
_str2bin["ipv6_nd_sll"] = mac_str2bin(OXM_OF_IPV6_ND_SLL)

_bin2str[OXM_OF_IPV6_ND_TLL] = mac_bin2str("ipv6_nd_tll")
_str2bin["ipv6_nd_tll"] = mac_str2bin(OXM_OF_IPV6_ND_TLL)

_bin2str[OXM_OF_MPLS_LABEL] = uint_bin2str("mpls_label={:#x}/{:#x}")
_str2bin["mpls_label"] = uint_str2bin(OXM_OF_MPLS_LABEL, 4)

_bin2str[OXM_OF_MPLS_TC] = uint_bin2str("mpls_tc={:d}")
_str2bin["mpls_tc"] = uint_str2bin(OXM_OF_MPLS_TC, 1)

_bin2str[OXM_OF_MPLS_BOS] = uint_bin2str("mpls_bos={:d}")
_str2bin["mpls_bos"] = uint_str2bin(OXM_OF_MPLS_BOS, 1)

_bin2str[OXM_OF_PBB_ISID] = uint_bin2str("pbb_isid={:#x}/{:#x}")
_str2bin["pbb_isid"] = uint_str2bin(OXM_OF_PBB_ISID, 3)

_bin2str[OXM_OF_TUNNEL_ID] = uint_bin2str("tunnel_id={:#x}/{:#x}")
_str2bin["tunnel_id"] = uint_str2bin(OXM_OF_TUNNEL_ID, 8)

_bin2str[OXM_OF_IPV6_EXTHDR] = uint_bin2str("ipv6_exthdr={:#x}/{:#x}")
_str2bin["ipv6_exthdr"] = uint_str2bin(OXM_OF_IPV6_EXTHDR, 2)

_bin2str[OXM_OF_PBB_UCA] = uint_bin2str("pbb_uca={:d}")
_str2bin["pbb_uca"] = uint_str2bin(OXM_OF_PBB_UCA, 1)

_bin2str[OXM_OF_TCP_FLAGS] = uint_bin2str("tcp_flags={:#04x}/{:#04x}")
_str2bin["tcp_flags"] = uint_str2bin(OXM_OF_TCP_FLAGS, 2)

_bin2str[OXM_OF_ACTSET_OUTPUT] = port_bin2str("actset_output")
_str2bin["actset_output"] = port_str2bin(OXM_OF_ACTSET_OUTPUT)

_bin2str[OXM_OF_PACKET_TYPE] = pkt_bin2str
_str2bin["packet_type"] = pkt_str2bin

_bin2str[NXM_OF_IN_PORT] = port_bin2str("nxm_in_port", fmt="!H")
_str2bin["nxm_in_port"] = port_str2bin(NXM_OF_IN_PORT, fmt="!H")

_bin2str[NXM_OF_ETH_DST] = mac_bin2str("nxm_eth_dst")
_str2bin["nxm_eth_dst"] = mac_str2bin(NXM_OF_ETH_DST)

_bin2str[NXM_OF_ETH_SRC] = mac_bin2str("nxm_eth_src")
_str2bin["nxm_eth_src"] = mac_str2bin(NXM_OF_ETH_SRC)

_bin2str[NXM_OF_ETH_TYPE] = uint_bin2str("nxm_eth_type={:#06x}/{:#06x}")
_str2bin["nxm_eth_type"] = uint_str2bin(NXM_OF_ETH_TYPE, 2)

_bin2str[NXM_OF_VLAN_TCI] = uint_bin2str("nxm_vlan_tci={:#x}/{:#06x}")
_str2bin["nxm_vlan_tci"] = uint_str2bin(NXM_OF_VLAN_TCI, 2)

_bin2str[NXM_OF_IP_TOS] = uint_bin2str("nxm_ip_tos={:#x}/{:#x}")
_str2bin["nxm_ip_tos"] = uint_str2bin(NXM_OF_IP_TOS, 1)

_bin2str[NXM_OF_IP_PROTO] = uint_bin2str("nxm_ip_proto={:d}/{:d}")
_str2bin["nxm_ip_proto"] = uint_str2bin(NXM_OF_IP_PROTO, 1)

_bin2str[NXM_OF_IP_SRC] = ipv4_bin2str("nxm_ip_src")
_str2bin["nxm_ip_src"] = ipv4_str2bin(NXM_OF_IP_SRC)

_bin2str[NXM_OF_IP_DST] = ipv4_bin2str("nxm_ip_dst")
_str2bin["nxm_ip_dst"] = ipv4_str2bin(NXM_OF_IP_DST)

_bin2str[NXM_OF_TCP_SRC] = uint_bin2str("nxm_tcp_src={:d}/{:d}")
_str2bin["nxm_tcp_src"] = uint_str2bin(NXM_OF_TCP_SRC, 2)

_bin2str[NXM_OF_TCP_DST] = uint_bin2str("nxm_tcp_dst={:d}/{:d}")
_str2bin["nxm_tcp_dst"] = uint_str2bin(NXM_OF_TCP_DST, 2)

_bin2str[NXM_OF_UDP_SRC] = uint_bin2str("nxm_udp_src={:d}/{:d}")
_str2bin["nxm_udp_src"] = uint_str2bin(NXM_OF_UDP_SRC, 2)

_bin2str[NXM_OF_UDP_DST] = uint_bin2str("nxm_udp_dst={:d}/{:d}")
_str2bin["nxm_udp_dst"] = uint_str2bin(NXM_OF_UDP_DST, 2)

_bin2str[NXM_OF_ICMP_TYPE] = uint_bin2str("nxm_icmp_type={:d}/{:d}")
_str2bin["nxm_icmp_type"] = uint_str2bin(NXM_OF_ICMP_TYPE, 1)

_bin2str[NXM_OF_ICMP_CODE] = uint_bin2str("nxm_icmp_code={:d}/{:d}")
_str2bin["nxm_icmp_code"] = uint_str2bin(NXM_OF_ICMP_CODE, 1)

_bin2str[NXM_OF_ARP_OP] = uint_bin2str("nxm_arp_op={:#d}/{:#d}")
_str2bin["nxm_arp_op"] = uint_str2bin(NXM_OF_ARP_OP, 2)

_bin2str[NXM_OF_ARP_SPA] = ipv4_bin2str("nxm_arp_spa")
_str2bin["nxm_arp_spa"] = ipv4_str2bin(NXM_OF_ARP_SPA)

_bin2str[NXM_OF_ARP_TPA] = ipv4_bin2str("nxm_arp_tpa")
_str2bin["nxm_arp_tpa"] = ipv4_str2bin(NXM_OF_ARP_TPA)

_bin2str[NXM_NX_REG0] = uint_bin2str("nxm_reg0={:#x}/{:#x}")
_str2bin["nxm_reg0"] = uint_str2bin(NXM_NX_REG0, 4)

_bin2str[NXM_NX_REG1] = uint_bin2str("nxm_reg1={:#x}/{:#x}")
_str2bin["nxm_reg1"] = uint_str2bin(NXM_NX_REG1, 4)

_bin2str[NXM_NX_REG2] = uint_bin2str("nxm_reg2={:#x}/{:#x}")
_str2bin["nxm_reg2"] = uint_str2bin(NXM_NX_REG2, 4)

_bin2str[NXM_NX_REG3] = uint_bin2str("nxm_reg3={:#x}/{:#x}")
_str2bin["nxm_reg3"] = uint_str2bin(NXM_NX_REG3, 4)

_bin2str[NXM_NX_REG4] = uint_bin2str("nxm_reg4={:#x}/{:#x}")
_str2bin["nxm_reg4"] = uint_str2bin(NXM_NX_REG4, 4)

_bin2str[NXM_NX_REG5] = uint_bin2str("nxm_reg5={:#x}/{:#x}")
_str2bin["nxm_reg5"] = uint_str2bin(NXM_NX_REG5, 4)

_bin2str[NXM_NX_REG6] = uint_bin2str("nxm_reg6={:#x}/{:#x}")
_str2bin["nxm_reg6"] = uint_str2bin(NXM_NX_REG6, 4)

_bin2str[NXM_NX_REG7] = uint_bin2str("nxm_reg7={:#x}/{:#x}")
_str2bin["nxm_reg7"] = uint_str2bin(NXM_NX_REG7, 4)

_bin2str[NXM_NX_TUN_ID] = uint_bin2str("nxm_tun_id={:#x}/{:#x}")
_str2bin["nxm_tun_id"] = uint_str2bin(NXM_NX_TUN_ID, 8)

_bin2str[NXM_NX_ARP_SHA] = mac_bin2str("nxm_arp_sha")
_str2bin["nxm_arp_sha"] = mac_str2bin(NXM_NX_ARP_SHA)

_bin2str[NXM_NX_ARP_THA] = mac_bin2str("nxm_arp_tha")
_str2bin["nxm_arp_tha"] = mac_str2bin(NXM_NX_ARP_THA)

_bin2str[NXM_NX_IPV6_SRC] = ipv6_bin2str("nxm_ipv6_src")
_str2bin["nxm_ipv6_src"] = ipv6_str2bin(NXM_NX_IPV6_SRC)

_bin2str[NXM_NX_IPV6_DST] = ipv6_bin2str("nxm_ipv6_dst")
_str2bin["nxm_ipv6_dst"] = ipv6_str2bin(NXM_NX_IPV6_DST)

_bin2str[NXM_NX_ICMPV6_TYPE] = uint_bin2str("nxm_icmpv6_type={:d}/{:d}")
_str2bin["nxm_icmpv6_type"] = uint_str2bin(NXM_NX_ICMPV6_TYPE, 1)

_bin2str[NXM_NX_ICMPV6_CODE] = uint_bin2str("nxm_icmpv6_code={:d}/{:d}")
_str2bin["nxm_icmpv6_code"] = uint_str2bin(NXM_NX_ICMPV6_CODE, 1)

_bin2str[NXM_NX_ND_TARGET] = ipv6_bin2str("nxm_nd_target")
_str2bin["nxm_nd_target"] = ipv6_str2bin(NXM_NX_ND_TARGET)

_bin2str[NXM_NX_ND_SLL] = mac_bin2str("nxm_nd_sll")
_str2bin["nxm_nd_sll"] = mac_str2bin(NXM_NX_ND_SLL)

_bin2str[NXM_NX_ND_TLL] = mac_bin2str("nxm_nd_tll")
_str2bin["nxm_nd_tll"] = mac_str2bin(NXM_NX_ND_TLL)

_bin2str[NXM_NX_IP_FRAG] = uint_bin2str("nxm_icmp_type={:d}/{:d}")
_str2bin["nxm_ip_frag"] = uint_str2bin(NXM_NX_IP_FRAG, 1)

_bin2str[NXM_NX_IPV6_LABEL] = uint_bin2str("nxm_ipv6_label={:#x}/{:#x}")
_str2bin["nxm_ipv6_label"] = uint_str2bin(NXM_NX_IPV6_LABEL, 4)

_bin2str[NXM_NX_IP_ECN] = uint_bin2str("nxm_ip_ecn={:#x}/{:#x}")
_str2bin["nxm_ip_ecn"] = uint_str2bin(NXM_NX_IP_ECN, 1)

_bin2str[NXM_NX_IP_TTL] = uint_bin2str("nxm_ip_ttl={:d}/{:d}")
_str2bin["nxm_ip_ttl"] = uint_str2bin(NXM_NX_IP_TTL, 1)

_bin2str[NXM_NX_TUN_IPV4_SRC] = ipv4_bin2str("nxm_tun_ipv4_src")
_str2bin["nxm_tun_ipv4_src"] = ipv4_str2bin(NXM_NX_TUN_IPV4_SRC)

_bin2str[NXM_NX_TUN_IPV4_DST] = ipv4_bin2str("nxm_tun_ipv4_dst")
_str2bin["nxm_tun_ipv4_dst"] = ipv4_str2bin(NXM_NX_TUN_IPV4_DST)

_bin2str[NXM_NX_PKT_MARK] = uint_bin2str("nxm_pkt_mark={:#x}/{:#x}")
_str2bin["nxm_pkt_mark"] = uint_str2bin(NXM_NX_PKT_MARK, 4)

_bin2str[NXM_NX_TCP_FLAGS] = uint_bin2str("nxm_tcp_flags={:#06x}/{:#06x}")
_str2bin["nxm_tcp_flags"] = uint_str2bin(NXM_NX_TCP_FLAGS, 2)

_bin2str[NXM_NX_DP_HASH] = uint_bin2str("nxm_dp_hash={:#010x}/{:#010x}")
_str2bin["nxm_dp_hash"] = uint_str2bin(NXM_NX_DP_HASH, 4)

_bin2str[NXM_NX_RECIRC_ID] = uint_bin2str("nxm_recirc_id={:#010x}/{:#010x}")
_str2bin["nxm_recirc_id"] = uint_str2bin(NXM_NX_RECIRC_ID, 4)

_bin2str[NXM_NX_CONJ_ID] = uint_bin2str("nxm_conj_id={:#010x}/{:#010x}")
_str2bin["nxm_conj_id"] = uint_str2bin(NXM_NX_CONJ_ID, 4)

_bin2str[NXM_NX_TUN_GBP_ID] = uint_bin2str("nxm_tun_gbp_id={:#06x}/{:#06x}")
_str2bin["nxm_tun_gbp_id"] = uint_str2bin(NXM_NX_TUN_GBP_ID, 2)

_bin2str[NXM_NX_TUN_GBP_FLAGS] = uint_bin2str("nxm_tun_gbp_flags={:#x}/{:#x}")
_str2bin["nxm_tun_gbp_flags"] = uint_str2bin(NXM_NX_TUN_GBP_FLAGS, 1)

_bin2str[STROXM_BASIC_DOT11] = uint_bin2str("dot11={:d}")
_str2bin["dot11"] = uint_str2bin(STROXM_BASIC_DOT11, 1)

_bin2str[STROXM_BASIC_DOT11_FRAME_CTRL] = hex_bin2str("dot11_frame_ctrl")
_str2bin["dot11_frame_ctrl"] = hex_str2bin(STROXM_BASIC_DOT11_FRAME_CTRL)

_bin2str[STROXM_BASIC_DOT11_ADDR1] = mac_bin2str("dot11_addr1")
_str2bin["dot11_addr1"] = mac_str2bin(STROXM_BASIC_DOT11_ADDR1)

_bin2str[STROXM_BASIC_DOT11_ADDR2] = mac_bin2str("dot11_addr2")
_str2bin["dot11_addr2"] = mac_str2bin(STROXM_BASIC_DOT11_ADDR2)

_bin2str[STROXM_BASIC_DOT11_ADDR3] = mac_bin2str("dot11_addr3")
_str2bin["dot11_addr3"] = mac_str2bin(STROXM_BASIC_DOT11_ADDR3)

_bin2str[STROXM_BASIC_DOT11_ADDR4] = mac_bin2str("dot11_addr4")
_str2bin["dot11_addr4"] = mac_str2bin(STROXM_BASIC_DOT11_ADDR4)

_bin2str[STROXM_BASIC_DOT11_SSID] = ssid_bin2str
_str2bin["dot11_ssid"] = ssid_str2bin

_bin2str[STROXM_BASIC_DOT11_ACTION_CATEGORY] = hex_bin2str("dot11_action_category")
_str2bin["dot11_action_category"] = hex_str2bin(STROXM_BASIC_DOT11_ACTION_CATEGORY)

_bin2str[STROXM_BASIC_DOT11_PUBLIC_ACTION] = uint_bin2str("dot11_public_action={:d}")
_str2bin["dot11_public_action"] = uint_str2bin(STROXM_BASIC_DOT11_PUBLIC_ACTION, 1)

_bin2str[STROXM_BASIC_DOT11_TAG] = uint_bin2str("dot11_tag={:d}")
_str2bin["dot11_tag"] = uint_str2bin(STROXM_BASIC_DOT11_TAG, 1)

_bin2str[STROXM_BASIC_DOT11_TAG_VENDOR] = hex_bin2str("dot11_tag_vendor")
_str2bin["dot11_tag_vendor"] = hex_str2bin(STROXM_BASIC_DOT11_TAG_VENDOR)

_bin2str[STROXM_RADIOTAP_TSFT] = le_bin2str("radiotap_tsft={:d}")
_str2bin["radiotap_tsft"] = le_str2bin(STROXM_RADIOTAP_TSFT, 8)

_bin2str[STROXM_RADIOTAP_FLAGS] = le_bin2str("radiotap_flags={:#04x}/{:#04x}")
_str2bin["radiotap_flags"] = le_str2bin(STROXM_RADIOTAP_FLAGS, 1)

_bin2str[STROXM_RADIOTAP_RATE] = rate_bin2str
_str2bin["radiotap_rate"] = rate_str2bin

_bin2str[STROXM_RADIOTAP_CHANNEL] = ch_bin2str
_str2bin["radiotap_channel"] = ch_str2bin

_bin2str[STROXM_RADIOTAP_FHSS] = hex_bin2str("radiotap_fhss")
_str2bin["radiotap_fhss"] = hex_str2bin(STROXM_RADIOTAP_FHSS)

_bin2str[STROXM_RADIOTAP_DBM_ANTSIGNAL] = s8_bin2str("radiotap_dbm_antsignal")
_str2bin["radiotap_dbm_antsignal"] = s8_str2bin(STROXM_RADIOTAP_DBM_ANTSIGNAL)

_bin2str[STROXM_RADIOTAP_DBM_ANTNOISE] = s8_bin2str("radiotap_dbm_antnoise")
_str2bin["radiotap_dbm_antnoise"] = s8_str2bin(STROXM_RADIOTAP_DBM_ANTNOISE)

_bin2str[STROXM_RADIOTAP_LOCK_QUALITY] = le_bin2str("radiotap_lock_quality={:d}")
_str2bin["radiotap_lock_quality"] = le_str2bin(STROXM_RADIOTAP_LOCK_QUALITY, 2)

_bin2str[STROXM_RADIOTAP_TX_ATTENUATION] = le_bin2str("radiotap_tx_attenuation={:d}")
_str2bin["radiotap_tx_attenuation"] = le_str2bin(STROXM_RADIOTAP_TX_ATTENUATION, 2)

_bin2str[STROXM_RADIOTAP_DB_TX_ATTENUATION] = le_bin2str("radiotap_db_tx_attenuation={:d}")
_str2bin["radiotap_db_tx_attenuation"] = le_str2bin(STROXM_RADIOTAP_DB_TX_ATTENUATION, 2)

_bin2str[STROXM_RADIOTAP_DBM_TX_POWER] = s8_bin2str("radiotap_dbm_tx_power")
_str2bin["radiotap_dbm_tx_power"] = s8_str2bin(STROXM_RADIOTAP_DBM_TX_POWER)

_bin2str[STROXM_RADIOTAP_ANTENNA] = le_bin2str("radiotap_antenna={:d}")
_str2bin["radiotap_antenna"] = le_str2bin(STROXM_RADIOTAP_ANTENNA, 1)

_bin2str[STROXM_RADIOTAP_DB_ANTSIGNAL] = le_bin2str("radiotap_db_antsignal={:d}")
_str2bin["radiotap_db_antsignal"] = le_str2bin(STROXM_RADIOTAP_DB_ANTSIGNAL, 1)

_bin2str[STROXM_RADIOTAP_DB_ANTNOISE] = le_bin2str("radiotap_db_antnoise={:d}")
_str2bin["radiotap_db_antnoise"] = le_str2bin(STROXM_RADIOTAP_DB_ANTNOISE, 1)

_bin2str[STROXM_RADIOTAP_RX_FLAGS] = le_bin2str("radiotap_rx_flags={:#06x}")
_str2bin["radiotap_rx_flags"] = le_str2bin(STROXM_RADIOTAP_RX_FLAGS, 2)

_bin2str[STROXM_RADIOTAP_TX_FLAGS] = le_bin2str("radiotap_tx_flags={:#06x}")
_str2bin["radiotap_tx_flags"] = le_str2bin(STROXM_RADIOTAP_TX_FLAGS, 2)

_bin2str[STROXM_RADIOTAP_RTS_RETRIES] = le_bin2str("radiotap_rts_retries={:d}")
_str2bin["radiotap_rts_retries"] = le_str2bin(STROXM_RADIOTAP_RTS_RETRIES, 1)

_bin2str[STROXM_RADIOTAP_DATA_RETRIES] = le_bin2str("radiotap_data_retries={:d}")
_str2bin["radiotap_data_retries"] = le_str2bin(STROXM_RADIOTAP_DATA_RETRIES, 1)

_bin2str[STROXM_RADIOTAP_MCS] = comp_bin2str("radiotap_mcs", "<3B", "{:#04x} {:d} {:d}".split())
_str2bin["radiotap_mcs"] = comp_str2bin(STROXM_RADIOTAP_MCS, "<3B")

_bin2str[STROXM_RADIOTAP_AMPDU_STATUS] = comp_bin2str("radiotap_ampdu_status", "<IHBB", "{:#010x} {:#06x} {:#04x} {:#02x}".split())
_str2bin["radiotap_ampdu_status"] = comp_str2bin(STROXM_RADIOTAP_AMPDU_STATUS, "<IHBB")

_bin2str[STROXM_RADIOTAP_VHT] = vht_bin2str
_str2bin["radiotap_vht"] = vht_str2bin


def oxm2str(msg, loop=True):
	tokens = []
	while len(msg) >= 4:
		(kls,f1,l) = struct.unpack_from("!HBB", msg)
		if kls == OFPXMC_EXPERIMENTER:
			exp = struct.unpack_from("!I", msg, 4)[0]
			if exp == STRATOS_EXPERIMENTER_ID:
				tokens.append(_bin2str[stratos(f1>>1)](msg[8:4+l], (f1&1)==1))
			else:
				tokens.append("?") # unknown experimenter id
		elif kls == OFPXMC_OPENFLOW_BASIC:
			tokens.append(_bin2str[f1>>1](msg[4:4+l], (f1&1)==1))
		elif kls == OFPXMC_NXM_0:
			tokens.append(_bin2str[nxm_of(f1>>1)](msg[4:4+l], (f1&1)==1))
		elif kls == OFPXMC_NXM_1:
			tokens.append(_bin2str[nxm_nx(f1>>1)](msg[4:4+l], (f1&1)==1))
		else:
			tokens.append("?") # unknown oxm class {:x}".format(kls)
		
		if loop:
			msg = msg[4+l:]
		else:
			break
	
	return ",".join(tokens)

def oxmid2str(msg, loop=True):
	tokens = []
	while len(msg) >= 4:
		rlen = 4
		(kls,f1,l) = struct.unpack_from("!HBB", msg)
		if kls == OFPXMC_EXPERIMENTER:
			rlen = 8
			exp = struct.unpack_from("!I", msg, 4)[0]
			if exp == STRATOS_EXPERIMENTER_ID:
				tokens.append(_bin2str[stratos(f1>>1)](None, (f1&1)==1))
			else:
				tokens.append("?") # unknown experimenter id
		elif kls == OFPXMC_OPENFLOW_BASIC:
			tokens.append(_bin2str[f1>>1](None, (f1&1)==1))
		elif kls == OFPXMC_NXM_0:
			tokens.append(_bin2str[nxm_of(f1>>1)](None, (f1&1)==1))
		elif kls == OFPXMC_NXM_1:
			tokens.append(_bin2str[nxm_nx(f1>>1)](None, (f1&1)==1))
		else:
			tokens.append("?") # unknown oxm class {:x}".format(kls)
		
		if loop:
			msg = msg[rlen:]
		else:
			break
	
	return ",".join(tokens)


def str2oxm(unparsed, loop=True):
	total = len(unparsed)
	msg = b""
	while unparsed:
		head,name,unparsed = get_token(unparsed)
		op,payload,unparsed = get_token(unparsed)
		assert head.find("=") < 0 and op.find("=") >= 0
		bin,length = _str2bin[name](payload)
		if not length:
			break
		
		assert length == len(payload), "{0} pos {1}".format(payload, length)
		msg += bin
		
		if not loop:
			break
	
	return msg, total - len(unparsed)


def str2oxmid(unparsed, loop=True, has_mask=True):
	ret = b""
	while unparsed:
		oxmid = None
		head,name,unparsed = get_token(unparsed)
		if name == "_":
			pass
		elif name in _str2bin:
			bin,length = _str2bin[name](has_mask)
			assert length == 0
			ret += bin
	
	return ret




#!python
import socket
import struct
import argparse
import re
import struct
from . import ofp4

class Eparam(Exception):
	pass

class Result(Exception):
	pass

def msgs(sock):
	while True:
		h = sock.recv(8)
		hi = struct.unpack("!BBHI", h)
		if hi[2] > 8:
			yield h+sock.recv(hi[2]-8)
		else:
			yield h

def ofctl():
	argp = argparse.ArgumentParser(description="ofpstr ofctl cli")
	argp.add_argument("switch", help="tcp:127.0.0.1:6633 or unix:/var/run/openvswitch/br0.mgmt")
	argp.add_argument("command", help="dump-flows, add-flow")
	argp.add_argument("args", nargs="*", help="dump-flows takes nothing, add-flow takes a flow rule")

	opts = argp.parse_args()
	sw = re.match(r"(tcp:(\d+.\d+.\d+.\d+):(\d+)|unix:(.*))", opts.switch)
	if not sw:
		raise Eparam("switch arg format error {0}".format(opts.switch))
	
	if sw.group(0).startswith("tcp:"):
		s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
		s.connect((sw.group(2), int(sw.group(3))))
	elif sw.group(0).startswith("unix:"):
		s = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
		s.connect(sw.group(4))
	
	s.send(struct.pack("!BBHI", 4, 0, 8, 1))
	if opts.command == "dump-flows":
		s.send(ofp4.str2flows([], type=18, xid=2)[0])
		for msg in msgs(s):
			if len(msg) < 16:
				continue
			mphdr = struct.unpack_from("!BBHIHH4x", msg)
			if mphdr[1] != 19 or mphdr[4] != 1: # OFPT_MULTIPART_REPLY, OFPMP_FLOW
				continue
			
			for rule in ofp4.flows2str(msg):
				print(rule)
			
			if mphdr[5]==0:
				break
	elif opts.command == "add-flow":
		x = ofp4.str2mod(" ".join(opts.args), xid=2)
		s.send(x)
		s.send(struct.pack("!BBHI", 4, 20, 8, 3))
		for msg in msgs(s):
			hdr = struct.unpack_from("!BBHI", msg)
			if hdr[1] == 1 and hdr[2] == 2:
				raise Result("failed")
			elif hdr[1] == 21 and hdr[3] == 3:
				break
	else:
		raise Eparam("unknown command")

if __name__=="__main__":
	ofctl()



try:
	from setuptools import setup
except:
	from distutils.core import setup

setup(name='ofpstr',
	version='0.1.12',
	description='Openflow stringer library',
	long_description=open("README.rst").read(),
	author='Hiroaki Kawai',
	author_email='hiroaki.kawai@gmail.com',
	url='https://github.com/hkwi/ofpstr/',
	packages=['ofpstr'],
	entry_points = {
		"console_scripts": ["ofpstr_ctl=ofpstr.cli:ofctl"],
	}
)



import unittest
import binascii
import ofpstr.oxm

class TestRoundTrip(unittest.TestCase):
	rules = (
		"in_port=any",
		"in_port=10",
		"in_phy_port=10",
		"metadata=0x5/0xff",
		"eth_src=00:00:00:00:00:00",
		"eth_src=00:00:00:00:00:00/01:00:00:00:00:00",
		"ipv4_src=192.168.0.1",
		"ipv4_src=192.168.0.1/255.255.255.0",
		"ipv4_src=192.168.0.1/255.0.255.255",
		"ipv6_src=::/ffff::",
		"vlan_vid=0x5",
		"vlan_vid=0x1000/0x1000",
		"pbb_isid=0x5",
		"packet_type=0x2:0x3",
		"dot11=0",
		"dot11=1",
		"dot11_frame_ctrl=0000",
		"dot11_frame_ctrl=0000/fff0",
		"dot11_addr1=ff:ff:ff:ff:ff:ff",
		"dot11_addr1=01:00:00:00:00:00/01:00:00:00:00:00", # broadcast,multicast
		"dot11_ssid=stratos1",
		"dot11_ssid=stratos/ffffffffffffff00000000",
		"dot11_action_category=03",
		"dot11_action_category=7f00e04d", # vendor action
		"dot11_public_action=10",         # GAS initial
		"dot11_tag=0",
		"dot11_tag_vendor=00e04d",
		"radiotap_tsft=1",
		"radiotap_flags=0x00",
		"radiotap_rate=500.0K",
		"radiotap_rate=11.0M",
		"radiotap_channel=2412:0x1234",
		"radiotap_channel=2412:0x1234/:0x00ff",
		"radiotap_fhss=0102",
		"radiotap_dbm_antsignal=-80",
		"radiotap_mcs=0x01:1:2",
		"radiotap_ampdu_status=0x12345678:0x1234:0x12:0x12",
		"radiotap_ampdu_status=:::/0x12345678:0x1234:0x12:0x12",
		"radiotap_vht=0x1234:0x12:0:12345678:0x12:0:0x0000",
		"nxm_in_port=any",
		"nxm_eth_dst=00:11:22:33:44:55",
		"nxm_eth_src=00:11:22:33:44:55/01:00:00:00:00:00",
		"nxm_eth_type=0x88a8",
		"nxm_vlan_tci=0x3001",
		"nxm_ip_tos=0x10", # IPTOS_LOWDELAY
		"nxm_ip_proto=6",
		"nxm_ip_src=192.168.0.1",
		"nxm_ip_dst=192.168.0.0/255.255.255.0",
		"nxm_tcp_src=80",
		"nxm_tcp_dst=8080",
		"nxm_udp_src=53",
		"nxm_udp_dst=4789",
		"nxm_icmp_type=30",
		"nxm_icmp_code=0",
		"nxm_arp_op=1",
		"nxm_arp_spa=192.168.0.1",
		"nxm_arp_tpa=192.168.0.0/255.255.255.0",
		)
	def test_all(self):
		for rule in self.rules:
			msg, length = ofpstr.oxm.str2oxm(rule)
			assert length == len(rule), "length {0}!={1} {2}".format(length, len(rule), rule)
			assert rule == ofpstr.oxm.oxm2str(msg), "{0}!={1} {2}".format(rule, ofpstr.oxm.oxm2str(msg), binascii.b2a_hex(msg))
	
	def test_id(self):
		for rule in self.rules:
			strid = rule.split("=")[0]
			oxmid = ofpstr.oxm.str2oxmid(strid)
			assert len(oxmid) in (4, 8)
			assert strid == ofpstr.oxm.oxmid2str(oxmid), "{0}!={1}".format(strid, ofpstr.oxm.oxmid2str(oxmid))

if __name__ == "__main__":
	unittest.main()



import binascii
import struct
from .util import ofpp, parseInt, get_token, split, parse_func, int2bytes
from .oxm import str2oxm, oxm2str, str2oxmid, oxmid2str

try:
	inttypes = (int, long)
except:
	inttypes = (int, )

align8 = lambda x:(x+7)//8*8

OFPAT_EXPERIMENTER = 0xffff
NX_VENDOR_ID = 0x00002320

nx_hash_names = ("eth_src", "symmetric_l4")
for n,name in enumerate(nx_hash_names):
	globals()["NX_HASH_FIELDS_{:s}".format(name.upper())] = n

nx_mp_alg_names = ("modulo_n", "hash_threshold", "hrw", "iter_hash")
for n,name in enumerate(nx_mp_alg_names):
	globals()["NX_MP_ALG_{:s}".format(name.upper())] = n

nx_bd_alg_names = ("active_backup", "hrw")
for n,name in enumerate(nx_bd_alg_names):
	globals()["NX_BD_ALG_{:s}".format(name.upper())] = n

class nxast(int):
	def __hash__(self):
		return hash((NX_VENDOR_ID, int(self)))

nxast_names = {
	21: "cnt_ids", # DEC_TTL_CNT_IDS; controller ids to packet-in
	7: "reg_load",
	33: "reg_load2",
	6: "reg_move",
	1: "resubmit",
	14: "resubmit_table",
	2: "set_tunnel",
	9: "set_tunnel64",
	5: "pop_queue",
	8: "note",
	10: "multipath",
	12: "bundle",
	13: "bundle_load",
	15: "output_reg",
	32: "output_reg2",
	16: "learn",
	17: "exit",
	19: "fin_timeout",
	20: "controller",
	22: "write_metadata",
	27: "stack_push",
	28: "stack_pop",
	29: "sample",
	34: "conjunction"
}
for n,name in nxast_names.items():
	globals()["NXAST_{:s}".format(name.upper())] = nxast(n)

def _pad8(bin):
	return bin+bytes(bytearray(align8(len(bin))-len(bin)))

def _nxast_hdr(subtype):
	return struct.pack("!HHIH", OFPAT_EXPERIMENTER, 10, NX_VENDOR_ID, subtype)

def _fix_act_len(bin):
	if len(bin) == struct.unpack_from("!H", bin, 2):
		return bin
	return bin[0:2]+struct.pack("!H", len(bin))+bin[4:]

def _nxast_str2act(func):
	def str2act(unparsed, readarg):
		bin,rlen = func(unparsed, readarg)
		return _fix_act_len(_pad8(bin)), rlen
	return str2act

def get_bits(unparsed):
	h1,F,unparsed = get_token(unparsed)
	name,arg = parse_func(F)
	oxmid = str2oxmid(name, has_mask=False, loop=False)
	oxm_class,oxm_field,oxm_length = struct.unpack_from("!HBB", oxmid)
	size = oxm_length
	if oxm_class == 0xffff:
		size = oxm_length-4
	
	shift = 0
	nbits = size*8
	if arg:
		shift,limit = map(lambda x: int(x) if len(x) else None, arg.split(":", 1))
		if shift is None:
			shift = 0
		if limit is None:
			limit = size*8
		nbits = limit - shift
	return oxmid, shift, nbits, unparsed

@_nxast_str2act
def action_cnt_ids_str2act(unparsed, readarg):
	if readarg:
		op,value,unparsed = get_token(unparsed)
		assert op.find("=")>=0
		ids = value.split(":")
		length = len(op)+len(value)
	else:
		ids = split(unparsed)
		length = 0
	ids = list(map(lambda x: parseInt(x)[0], ids))
	return _nxast_hdr(NXAST_CNT_IDS)+struct.pack("!H4x{:d}H".format(len(ids)), len(ids), *ids), length

def action_cnt_ids_act2str(payload):
	num = struct.unpack_from("!H", payload)[0]
	return "cnt_ids({:s})".format(",".join(map(lambda x: "{:#x}".format(x), struct.unpack_from("!{:d}H".format(num), payload, 6))))

def action_reg_load_str2act(subtype):
	@_nxast_str2act
	def str2act(unparsed, readarg):
		if readarg:
			raise ValueError("reg_load argument error")
		oxm, rlen = str2oxm(unparsed, loop=False)
		if rlen==0:
			raise ValueError("invalid reg_load argument {0}".format(unparsed))
		
		oxm_class, oxm_field, oxm_length = struct.unpack_from("!HBB", oxm)
		payload = oxm[4:]
		if oxm_class == 0xffff:
			payload = oxm[6:]
		
		if subtype == NXAST_REG_LOAD:
			shift = 0
			nbits = None
			if oxm_field&0x1: # has_mask
				size = len(payload)//2
				mask = payload[size:]
				payload = payload[:size]
				for v in reversed(bytearray(mask)):
					for d in range(8):
						if v & (1<<d):
							if nbits is None:
								nbits = 0
							nbits += 1
						elif nbits is None:
							shift += 1
						else:
							break
			else:
				size = len(payload)
				nbits = size * 8
			
			value = 0
			for p in bytearray(payload):
				value = (value<<8) + p
			
			value = value>>shift
			return _nxast_hdr(NXAST_REG_LOAD)+struct.pack("!HHBBQ",
				(shift<<6)+(nbits-1), oxm_class, oxm_field&0xfe, size, value), rlen
		elif subtype == NXAST_REG_LOAD2:
			return _nxast_hdr(NXAST_REG_LOAD2)+oxm, rlen
		else:
			return b"", 0
	return str2act

def action_reg_load_act2str(payload):
	ofs_nbits,oxm_class,oxm_field,oxm_length,value = struct.unpack_from("!HHBBQ", payload)
	name = oxmid2str(payload[2:6])
	oxmid = str2oxmid(name, loop=False, has_mask=False)
	info = struct.unpack_from("!HBB", oxmid)
	size = info[2]
	
	shift = ofs_nbits>>6
	nbits = (ofs_nbits & 0x3f) + 1
	if shift == 0 and nbits == size*8:
		has_mask = False
		u = int2bytes([(value>>(s*8))&0xff for s in reversed(range(size))])
		arg = oxm2str(oxmid+u)
	else:
		value = value<<shift
		mask = 0
		for s in range(nbits):
			mask = (mask<<1) + 1
		for s in range(shift):
			mask = mask<<1
		
		u = int2bytes([
			(value>>(s*8))&0xff for s in reversed(range(size))
			]+[
			(mask>>(s*8))&0xff for s in reversed(range(size))
			])
		arg = oxm2str(struct.pack("!HBB", oxm_class, oxm_field|0x1, size*2)+u)
	
	return "reg_load({:s})".format(arg)

def action_reg_load2_act2str(payload):
	return "reg_load2({:s})".format(oxm2str(payload, loop=False))

@_nxast_str2act
def action_reg_move_str2act(unparsed, readarg):
	if readarg:
		raise ValueError("reg_move argument error")
	
	total = len(unparsed)
	loxm, lshift, lnbits, unparsed = get_bits(unparsed)
	roxm, rshift, rnbits, unparsed = get_bits(unparsed)
	assert lnbits == rnbits
	
	return _nxast_hdr(NXAST_REG_MOVE)+struct.pack(
		"!HHH", lnbits, rshift, lshift)+roxm+loxm, total-len(unparsed)

def action_reg_move_act2str(payload):
	n_bits, src_ofs, dst_ofs = struct.unpack_from("!HHH", payload)
	src = oxmid2str(payload[6:], loop=False)
	sinfo = struct.unpack_from("!HBB", payload, 6)
	if src_ofs == 0 and n_bits == sinfo[2]*8:
		arg = src
	else:
		arg = "{:s}[{:d}:{:d}]".format(src, src_ofs, src_ofs+n_bits)
	
	doffset = 10
	if sinfo[0] == 0xffff:
		doffset = 14
	dst = oxmid2str(payload[doffset:], loop=False)
	dinfo = struct.unpack_from("!HBB", payload, doffset)
	if dst_ofs == 0 and n_bits == sinfo[2]*8:
		arg = dst+"="+arg
	else:
		arg = "{:s}[{:d}:{:d}]={:s}".format(dst, dst_ofs, dst_ofs+n_bits, arg)
	
	return "reg_move({:s})".format(arg)

def action_resubmit_str2act(subtype):
	@_nxast_str2act
	def str2act(unparsed, readarg):
		total = len(unparsed)
		h,p,unparsed=get_token(unparsed)
		port = None
		if p.upper() == "IN_PORT":
			port = 0xfff8
		if port is None:
			port, rlen = parseInt(p)
			assert len(p) == rlen
			assert port < 0xff00
		
		tbl = ""
		if readarg:
			assert "=" in h
			if subtype == NXAST_RESUBMIT_TABLE and unparsed.startswith(":"):
				h2,tbl,unparsed = get_token(unparsed[1:])
		elif subtype == NXAST_RESUBMIT_TABLE:
			h2,tbl,unparsed = get_token(unparsed)
		
		if len(tbl)==0 or tbl.upper() == "ALL":
			table = 0xff
		else:
			table, rlen = parseInt(tbl)
			assert rlen == len(tbl)
		
		return _nxast_hdr(subtype)+struct.pack("!HB", port, table), total-len(unparsed)
	return str2act

def action_resubmit_act2str(payload):
	p = struct.unpack_from("!H", payload)[0]
	port = "in_port" if p==0xfff8 else "{:d}".format(p)
	return "resubmit({:s})".format(port)

def action_resubmit_table_act2str(payload):
	p,tbl = struct.unpack_from("!HB", payload)
	port = "in_port" if p==0xfff8 else "{:d}".format(p)
	if tbl == 0xff:
		table = "all"
	else:
		table = "{:d}".format(tbl)
	return "resubmit_table({:s},{:s})".format(port, table)

def action_set_tunnel_str2act(subtype):
	@_nxast_str2act
	def str2act(unparsed, readarg):
		h,b,unparsed = get_token(unparsed)
		if readarg:
			assert "=" in h
		num, rlen = parseInt(b)
		assert rlen==len(b)
		fmt = "!I"
		stype = subtype
		if num > 0xffffffff:
			stype = NXAST_SET_TUNNEL64
			fmt = "!Q"
		
		return _nxast_hdr(stype)+struct.pack(fmt, num), len(h)+len(b)
	return str2act

def action_set_tunnel_act2str(size, fmt):
	def act2str(payload):
		if size==4:
			num = struct.unpack_from("!I", payload)[0]
		elif size==8:
			num = struct.unpack_from("!Q", payload)[0]
		return fmt.format(num)
	return act2str

def action_null_str2act(subtype):
	@_nxast_str2act
	def str2act(unparsed, readarg):
		return _nxast_hdr(subtype), 0
	return str2act

def action_null_act2str(fmt):
	def act2str(payload):
		return fmt
	return act2str

def action_qp_str2act(subtype):
	def str2act(unparsed, readarg):
		h,b,unparsed = get_token(unparsed)
		if readarg:
			assert "=" in h
		return _fix_act_len(_nxast_hdr(subtype)+binascii.a2b_qp(b)), len(h)+len(b)
	return str2act

def action_qp_act2str(fmt):
	def act2str(payload):
		return fmt.format(binascii.b2a_qp(payload).decode("UTF-8"))
	return act2str

@_nxast_str2act
def action_multipath_str2act(unparsed, readarg):
	if readarg:
		raise ValueError("multipath argument must be in func style")
	total = len(unparsed)
	h,hash_name,unparsed = get_token(unparsed)
	mhash = nx_hash_names.index(hash_name)
	h,basis_str,unparsed = get_token(unparsed)
	basis,rlen = parseInt(basis_str) # basis, nx_hash universal parameter
	assert rlen == len(basis_str)
	h,mp_alg_name,unparsed = get_token(unparsed)
	malg = nx_mp_alg_names.index(mp_alg_name)
	h,link_str,unparsed = get_token(unparsed)
	max_link, rlen = parseInt(link_str)
	assert rlen == len(link_str)
	h,arg_str,unparsed = get_token(unparsed)
	arg, rlen = parseInt(arg_str)
	h3,dst,unparsed = get_token(unparsed)
	oxm, shift, nbits, junk = get_bits(dst)
	assert len(junk) == 0
	return _nxast_hdr(NXAST_MULTIPATH)+struct.pack("!HH2xHHI2xH",
		mhash, basis, malg, max_link, arg, (shift<<6)+(nbits-1)
		)+oxm, total-len(unparsed)

def action_multipath_act2str(payload):
	mhash,basis,malg,max_link,arg,ofs_nbits = struct.unpack_from("!HH2xHHI2xH", payload)
	shift = ofs_nbits>>6
	nbits = (ofs_nbits&0x3f)+1
	dst = oxmid2str(payload[18:], loop=False)
	return "multipath({:s},{:d},{:s},{:d},{:#x},{:s}[{:d}:{:d}])".format(
		nx_hash_names[mhash],
		basis,
		nx_mp_alg_names[malg],
		max_link,
		arg,
		dst,
		shift,
		shift+nbits)

@_nxast_str2act
def action_bundle_str2act(unparsed, readarg):
	total = len(unparsed)
	if readarg:
		raise ValueError("func style")
	# bundle(<hash_fields>, <basis>, <bundle_algorithm>, <slave_type>, slaves(<slaves>, ...))
	_,fields,unparsed = get_token(unparsed)
	_,basis,unparsed = get_token(unparsed)
	_,algorithm,unparsed = get_token(unparsed)
	_,slave_type,unparsed = get_token(unparsed)
	_,slaves,unparsed = get_token(unparsed)
	
	fields = nx_hash_names.index(fields)
	basis, rlen = parseInt(basis)
	algorithm = nx_bd_alg_names.index(algorithm)
	oxmid = str2oxmid(slave_type, loop=False)
	sname, sargs = parse_func(slaves)
	assert sname == "slaves"
	slaves = []
	if sargs:
		for p in split(sargs):
			port = None
			for num,pname in ofpp.items():
				if pname == p:
					port = num
			if port is None:
				port,rlen = parseInt(p)
			slaves.append(port)
	
	parts = [_nxast_hdr(NXAST_BUNDLE),
		struct.pack("!HHH", algorithm, fields, basis),
		oxmid,
		struct.pack("!HHI4x", len(slaves), 0, 0),
		struct.pack("!{:d}H".format(len(slaves)), *slaves)]
	return b"".join(parts), total-len(unparsed)

def action_bundle_act2str(payload):
	# algorithm,fields,basis,slave_type,n_slaves,ofs_nbits,dst,zero[4]
	fs = "!HHHIHHI4x"
	f = struct.unpack_from(fs, payload)
	# bundle(<hash_fields>, <basis>, <bundle_algorithm>, <slave_type>(<slaves>, ...))
	ports = struct.unpack_from("!{:d}H".format(f[4]), payload, struct.calcsize(fs))
	slaves = []
	for p in ports:
		port = None
		for num,pname in ofpp.items():
			if (num & 0xffff) == p:
				port = pname
		
		if port is None:
			port = "{:d}".format(p)
		
		slaves.append(port)
	
	return "bundle({:s},{:d},{:s},{:s},slaves({:s}))".format(
		nx_hash_names[f[1]],
		f[2],
		nx_bd_alg_names[f[0]],
		oxmid2str(payload[6:10], loop=False),
		",".join(slaves),
		)

@_nxast_str2act
def action_bundle_load_str2act(unparsed, readarg):
	total = len(unparsed)
	if readarg:
		raise ValueError("func style")
	# bundle_load(<hash_fields>, <basis>, <bundle_algorithm>, <slave_type>, <dst>[start:end], slaves(1,2,3))
	_,fields,unparsed = get_token(unparsed)
	_,basis,unparsed = get_token(unparsed)
	_,algorithm,unparsed = get_token(unparsed)
	_,slave_type,unparsed = get_token(unparsed)
	_,dst,unparsed = get_token(unparsed)
	_,slave,unparsed = get_token(unparsed)
	
	fields = nx_hash_names.index(fields)
	basis, rlen = parseInt(basis)
	algorithm = nx_bd_alg_names.index(algorithm)
	dname, dargs = parse_func(dst)
	doxmid = str2oxmid(dname, loop=False)
	if dargs:
		r = list(map(int, dargs.split(":", 1)))
	else:
		cls,field,size=struct.unpack("!HBB", doxmid)
		r = [0, size*8]
	
	sname, sargs = parse_func(slave)
	assert sname == "slaves"
	slaves = []
	if sargs:
		for p in split(sargs):
			port = None
			for num,pname in ofpp.items():
				if pname == p:
					port = num
			if port is None:
				port,rlen = parseInt(p)
			slaves.append(port)
	
	parts = [_nxast_hdr(NXAST_BUNDLE_LOAD),
		struct.pack("!HHH", algorithm, fields, basis),
		str2oxmid(slave_type, loop=False),
		struct.pack("!HH", len(slaves), (r[0]<<6)+(r[1]-1)),
		doxmid,
		b"\0"*4, # padding
		struct.pack("!{:d}H".format(len(slaves)), *slaves)]
	return b"".join(parts), total-len(unparsed)

def action_bundle_load_act2str(payload):
	# algorithm,fields,basis,slave_type,n_slaves,ofs_nbits,dst,zero[4]
	fs = "!HHHIHHI4x"
	f = struct.unpack_from(fs, payload)
	shift = f[5]>>6
	nbits = (f[5]&0x3f)+1
	# bundle(<hash_fields>, <basis>, <bundle_algorithm>, <slave_type>(<slaves>, ...))
	ports = struct.unpack_from("!{:d}H".format(f[4]), payload, struct.calcsize(fs))
	slaves = []
	for p in ports:
		port = None
		for num,pname in ofpp.items():
			if (num & 0xffff) == p:
				port = pname
		
		if port is None:
			port = "{:d}".format(p)
		
		slaves.append(port)
	
	return "bundle_load({:s},{:d},{:s},{:s},{:s}[{:d}:{:d}],slaves({:s}))".format(
		nx_hash_names[f[1]],
		f[2],
		nx_bd_alg_names[f[0]],
		oxmid2str(payload[6:10], loop=False),
		oxmid2str(payload[14:18], loop=False),
		shift,
		shift+nbits,
		",".join(slaves),
		)

def action_output_reg_str2act(subtype):
	@_nxast_str2act
	def str2act(unparsed, readarg):
		total = len(unparsed)
		# output_reg(<src>[<start>:<end>])
		h,arg,unparsed = get_token(unparsed)
		name, arg = parse_func(arg)
		oxmid = str2oxmid(name, loop=False, has_mask=False)
		if arg:
			r = list(map(int, arg.split(":", 1)))
		else:
			cls,field,size = struct.unpack_from("!HBB", oxmid)
			r = [0, size*8]
		
		max_len = 0xffff
		h,arg,unparsed = get_token(unparsed)
		if len(arg) and max_len:
			max_len,rlen = parseInt(arg)
		
		stype = subtype
		if len(oxmid) > 4:
			stype = NXAST_OUTPUT_REG2
		
		if stype == NXAST_OUTPUT_REG:
			parts = [_nxast_hdr(stype),
				struct.pack("!H", (r[0]<<6)+((r[1]&0x3f)-1)),
				oxmid,
				struct.pack("!H", max_len),
				b"\0"*6]
			return b"".join(parts), total-len(unparsed)
		elif stype == NXAST_OUTPUT_REG2:
			parts = [_nxast_hdr(stype),
				struct.pack("!HH", (r[0]<<6)+((r[1]&0x3f)-1), max_len),
				oxmid]
			return b"".join(parts), total-len(unparsed)
	return str2act

def action_output_reg_act2str(payload):
	f = struct.unpack_from("!HHBBH6x", payload)
	shift = f[0]>>6
	nbits = (f[0]&0x3f)+1
	size = f[3]
	name = oxmid2str(payload[2:6], loop=False)
	if shift == 0 and nbits == size*8:
		return "output_reg({:s})".format(name)
	else:
		return "output_reg({:s}[{:d}:{:d}])".format(
			name, shift, shift+nbits)

def action_output_reg2_act2str(payload):
	f = struct.unpack_from("!HHHBB", payload)
	shift = f[0]>>6
	nbits = (f[0]&0x3f)+1
	size = f[4]
	name = oxmid2str(payload[4:], loop=False)
	if shift == 0 and nbits == size*8:
		return "output_reg2({:s})".format(name)
	else:
		return "output_reg2({:s}[{:d}:{:d}])".format(
			name, shift, shift+nbits)

fixed_keys = '''idle_timeout
	hard_timeout
	priority
	cookie
	flags
	table
	fin_idle_timeout
	fin_hard_timeout'''.split()

@_nxast_str2act
def action_learn_str2act(unparsed, readarg):
	total = len(unparsed)
	if readarg:
		raise ValueError("learn must use func style")
	
	SRC_SHIFT = 12
	DST_SHIFT = 10
	def flow_mod_spec(src, dst, mode):
		flags = 0
		nbits = None
		
		sname,sarg = parse_func(src)
		soxm = str2oxmid(sname, has_mask=False)
		if soxm:
			assert len(soxm) == 4
			_,_,size = struct.unpack("!HBB", soxm)
			sshift = 0
			if sarg:
				r = list(map(lambda n: int(n) if n else None, sarg.split(":", 1)))
				if r[0] is not None:
					sshift = r[0]
				if r[1] is None:
					nbits = size*8 - sshift
				else:
					nbits = r[1] - sshift
			else:
				nbits = size*8
			sobj = soxm+struct.pack("!H", sshift)
		else:
			sobj,_ = parseInt(sname)
			flags |= 1<<SRC_SHIFT
		
		if mode == "output":
			flags += 2<<DST_SHIFT
			if isinstance(sobj, inttypes):
				nbits = 2
				return struct.pack("!HH", flags+nbits, sobj)
			else:
				return struct.pack("!H", flags+nbits)+sobj
		
		dname,darg = parse_func(dst)
		doxm = str2oxmid(dname, has_mask=False)
		assert len(doxm) == 4
		_,_,size = struct.unpack("!HBB", doxm)
		dnbits = size*8
		dshift = 0
		if darg:
			r = list(map(lambda n: int(n) if n else None, darg.split(":", 1)))
			if r[0] is not None:
				dshift = r[0]
			if r[1] is not None:
				dnbits = r[1]-dshift
		if nbits:
			assert nbits == dnbits
		else:
			nbits = dnbits
		
		dbin = doxm+struct.pack("!H", dshift)
		if isinstance(sobj, inttypes):
			sobj = int2bytes([(sobj>>(s*8))&0xff for s in reversed(range((nbits+15)//16*2))])
		
		if mode == "reg_load":
			flags += 1<<DST_SHIFT
		return struct.pack("!H", flags+nbits)+sobj+dbin
	
	info = {}
	specs = b""
	left = None
	def single(token):
		fname,farg = parse_func(token)
		assert farg, left
		_,dst,farg = get_token(farg)
		h,src,farg = get_token(farg)
		assert "=" in h
		return flow_mod_spec(src, dst, fname)
	
	while unparsed:
		h,arg,unparsed = get_token(unparsed)
		if "=" in h:
			if left in fixed_keys:
				info[left],rlen = parseInt(arg)
			else:
				specs += flow_mod_spec(arg, left, None)
			left = None
		else:
			if left:
				specs += single(left)
			left = arg
	if left:
		specs += single(left)
	
	return _nxast_hdr(NXAST_LEARN)+struct.pack("!HHHQHBxHH",
		info.get("idle_timeout", 0),
		info.get("hard_timeout", 0),
		info.get("priority", 0),
		info.get("cookie", 0),
		info.get("flags", 0),
		info.get("table", 0),
		info.get("fin_idle_timeout", 0),
		info.get("fin_hard_timeout", 0))+specs, total-len(unparsed)

def action_learn_act2str(payload):
	ret = []
	
	fixed = "!HHHQHBxHH"
	f = struct.unpack_from(fixed, payload)
	if f[0]:
		ret.append("idle_timeout={:d}", f[0])
	if f[1]:
		ret.append("hard_timeout={:d}", f[1])
	if f[2]:
		ret.append("priority={:d}", f[2])
	if f[3]:
		ret.append("cookie={:#x}", f[3])
	if f[4]:
		ret.append("flags={:#x}", f[4])
	if f[5]:
		ret.append("table={:d}", f[5])
	if f[6]:
		ret.append("fin_idle_timeout={:d}", f[6])
	if f[7]:
		ret.append("fin_hard_timeout={:d}", f[7])
	
	payload = payload[struct.calcsize(fixed):]
	specs = []
	while len(payload)>2:
		n = struct.unpack_from("!H", payload)[0]
		if not n:
			break
		payload = payload[2:]
		
		nbits = n & 0x3ff
		dst = (n>>10) & 0x03
		src = (n>>12) & 0x01
		assert (n>>13) == 0
		if src:
			rlen = (nbits+15)//16*2
			n = 0
			for i in bytearray(payload[:rlen]):
				n = (n<<8) + i
			payload = payload[rlen:]
			sobj = n
		else:
			sname = oxmid2str(payload, loop=False)
			_,_,size,shift = struct.unpack_from("!HBBH", payload)
			payload = payload[6:]
			if shift==0 and size*8==nbits:
				sobj = sname
			else:
				sobj = "{:s}[{:d}:{:d}]".format(
					sname,
					shift,
					shift+nbits)
		
		if dst != 2:
			if isinstance(sobj, inttypes):
				sobj = "{:#x}".format(sobj)
			
			dname = oxmid2str(payload, loop=False)
			_,_,size,shift = struct.unpack_from("!HBBH", payload)
			payload = payload[6:]
			if shift==0 and size*8==nbits:
				dobj = dname
			else:
				dobj = "{:s}[{:d}:{:d}]".format(
					dname,
					shift,
					shift+nbits)
			if dst == 0:
				spec = "{:s}={:s}".format(dobj, sobj)
			elif dst == 1:
				spec = "reg_load({:s}={:s})".format(dobj, sobj)
		else:
			if isinstance(sobj, inttypes):
				sobj = "{:d}".format(sobj)
			
			spec= "output({:s})".format(sobj)
		specs.append(spec)
	
	ret.append("learn({:s})".format(",".join(specs)))
	return ",".join(ret)

_str2act = {}
_act2str = {}

def register_nxast(str2act, act2str):
	str2act.update(_str2act)
	act2str.update(_act2str)

#
# fin_timeout
# controller
# write_metadata
# stack_push
# stack_pop
# sample
# conjunction
#

_str2act["cnt_ids"] = action_cnt_ids_str2act
_act2str[NXAST_CNT_IDS] = action_cnt_ids_act2str

_str2act["reg_load"] = action_reg_load_str2act(NXAST_REG_LOAD)
_act2str[NXAST_REG_LOAD] = action_reg_load_act2str

_str2act["reg_load2"] = action_reg_load_str2act(NXAST_REG_LOAD2)
_act2str[NXAST_REG_LOAD2] = action_reg_load2_act2str

_str2act["reg_move"] = action_reg_move_str2act
_act2str[NXAST_REG_MOVE] = action_reg_move_act2str

_str2act["resubmit"] = action_resubmit_str2act(NXAST_RESUBMIT)
_act2str[NXAST_RESUBMIT] = action_resubmit_act2str

_str2act["resubmit_table"] = action_resubmit_str2act(NXAST_RESUBMIT_TABLE)
_act2str[NXAST_RESUBMIT_TABLE] = action_resubmit_table_act2str

_str2act["set_tunnel"] = action_set_tunnel_str2act(NXAST_SET_TUNNEL)
_act2str[NXAST_SET_TUNNEL] = action_set_tunnel_act2str(4, "set_tunnel({:#x})")

_str2act["set_tunnel64"] = action_set_tunnel_str2act(NXAST_SET_TUNNEL64)
_act2str[NXAST_SET_TUNNEL64] = action_set_tunnel_act2str(8, "set_tunnel64({:#x})")

_str2act["pop_queue"] = action_null_str2act(NXAST_POP_QUEUE)
_act2str[NXAST_POP_QUEUE] = action_null_act2str("pop_queue")

_str2act["note"] = action_qp_str2act(NXAST_NOTE)
_act2str[NXAST_NOTE] = action_qp_act2str("note({:s})")

_str2act["exit"] = action_null_str2act(NXAST_EXIT)
_act2str[NXAST_EXIT] = action_null_act2str("exit")

_str2act["multipath"] = action_multipath_str2act
_act2str[NXAST_MULTIPATH] = action_multipath_act2str

_str2act["bundle"] = action_bundle_str2act
_act2str[NXAST_BUNDLE] = action_bundle_act2str

_str2act["bundle_load"] = action_bundle_load_str2act
_act2str[NXAST_BUNDLE_LOAD] = action_bundle_load_act2str

_str2act["output_reg"] = action_output_reg_str2act(NXAST_OUTPUT_REG)
_act2str[NXAST_OUTPUT_REG] = action_output_reg_act2str

_str2act["output_reg2"] = action_output_reg_str2act(NXAST_OUTPUT_REG2)
_act2str[NXAST_OUTPUT_REG2] = action_output_reg2_act2str

_str2act["learn"] = action_learn_str2act
_act2str[NXAST_LEARN] = action_learn_act2str

